from spaceone.inventory.manager.change_history_manager import ChangeHistoryManager
from spaceone.inventory.manager.cleanup_manager import CleanupManager
from spaceone.inventory.manager.cloud_service_manager import CloudServiceManager
from spaceone.inventory.manager.cloud_service_type_manager import CloudServiceTypeManager
from spaceone.inventory.manager.cloud_service_query_set_manager import CloudServiceQuerySetManager
from spaceone.inventory.manager.cloud_service_stats_manager import CloudServiceStatsManager
from spaceone.inventory.manager.collecting_manager import CollectingManager
from spaceone.inventory.manager.collection_state_manager import CollectionStateManager
from spaceone.inventory.manager.collector_manager import CollectorManager
from spaceone.inventory.manager.collector_plugin_manager import CollectorPluginManager
from spaceone.inventory.manager.collector_rule_manager import CollectorRuleManager
from spaceone.inventory.manager.config_manager import ConfigManager
from spaceone.inventory.manager.identity_manager import IdentityManager
from spaceone.inventory.manager.job_manager import JobManager
from spaceone.inventory.manager.job_task_manager import JobTaskManager
from spaceone.inventory.manager.note_manager import NoteManager
from spaceone.inventory.manager.plugin_manager import PluginManager
from spaceone.inventory.manager.record_manager import RecordManager
from spaceone.inventory.manager.region_manager import RegionManager
from spaceone.inventory.manager.repository_manager import RepositoryManager
from spaceone.inventory.manager.resource_group_manager import ResourceGroupManager
from spaceone.inventory.manager.secret_manager import SecretManager



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel
from spaceone.inventory.model.collector_model import Collector


class Error(EmbeddedDocument):
    error_code = StringField(max_length=128)
    message = StringField(max_length=2048)
    additional = DictField()


class Job(MongoModel):
    job_id = StringField(max_length=40, generate_id='job', unique=True)
    status = StringField(max_length=20, default='IN_PROGRESS',
                         choices=('CANCELED', 'IN_PROGRESS', 'FAILURE', 'SUCCESS', 'ERROR'))
    # filters = DictField()                                                                     # Deprecated
    total_tasks = IntField(min_value=0, max_value=65000, default=0)
    remained_tasks = IntField(max_value=65000, default=0)
    success_tasks = IntField(min_value=0, max_value=65000, default=0)
    failure_tasks = IntField(min_value=0, max_value=65000, default=0)
    # errors = ListField(EmbeddedDocumentField(Error, default=None, null=True))                 # Deprecated
    collector = ReferenceField('Collector', reverse_delete_rule=CASCADE)
    collector_id = StringField(max_length=40)
    secret_id = StringField(max_length=40, null=True, default=None)
    plugin_id = StringField(max_length=40)
    # project_id = StringField(max_length=40, default=None, null=True)                          # Deprecated
    projects = ListField(StringField(max_length=40), default=[])
    domain_id = StringField(max_length=255)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)
    finished_at = DateTimeField(default=None, null=True)
    mark_error = IntField(min_value=0, default=0)

    meta = {
        'updatable_fields': [
            'status',
            'total_tasks',
            'remained_tasks',
            'success_tasks',
            'failure_tasks',
            # 'errors',
            'collector_id',
            'projects',
            'finished_at',
            'mark_error',
        ],
        'minimal_fields': [
            'job_id',
            'status',
            'created_at',
            'finished_at',
        ],
        'change_query_keys': {
            'user_projects': 'projects'
        },
        'reference_query_keys': {
            'collector': Collector
        },
        'ordering': [
            '-created_at'
        ],
        'indexes': [
            # 'job_id',
            'status',
            # 'collector',
            'collector_id',
            # 'project_id',
            # 'projects',
            'domain_id',
            'created_at',
            'finished_at'
        ]
    }




import logging
import copy
from spaceone.core.service import *
from spaceone.core import utils
from spaceone.inventory.error import *
from spaceone.inventory.manager.collection_state_manager import CollectionStateManager
from spaceone.inventory.manager.collector_manager import CollectorManager
from spaceone.inventory.manager.collector_plugin_manager import CollectorPluginManager
from spaceone.inventory.manager.collector_rule_manager import CollectorRuleManager
from spaceone.inventory.manager.repository_manager import RepositoryManager
from spaceone.inventory.manager.plugin_manager import PluginManager
from spaceone.inventory.manager.secret_manager import SecretManager
from spaceone.inventory.manager.job_manager import JobManager
from spaceone.inventory.manager.job_task_manager import JobTaskManager
from spaceone.inventory.manager.identity_manager import IdentityManager

_LOGGER = logging.getLogger(__name__)
_KEYWORD_FILTER = ['collector_id', 'name', 'provider']


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class CollectorService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['name', 'plugin_info', 'domain_id'])
    def create(self, params):
        """
        Args:
            params (dict): {
                'name': 'str',
                'plugin_info': 'dict',
                'provider': 'str',
                'schedule': 'dict',
                'secret_filter': 'dict',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            collector_vo (object)
        """
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        collector_plugin_mgr: CollectorPluginManager = self.locator.get_manager(CollectorPluginManager)
        plugin_manager = self.locator.get_manager(PluginManager)

        plugin_info = params['plugin_info']
        domain_id = params['domain_id']

        if 'tags' in params:
            params['tags'] = self._convert_tags(params.get('tags'))

        plugin_info_from_repository = self._get_plugin_from_repository(params['plugin_info'], params['domain_id'])

        params.update({
            'capability': plugin_info_from_repository.get('capability', {}),
            'provider': self._get_plugin_providers(params.get('provider'), plugin_info_from_repository)
        })

        if 'secret_filter' in params:
            self.validate_secret_filter(params['secret_filter'], domain_id)

        collector_vo = collector_mgr.create_collector(params)

        endpoint, updated_version = plugin_manager.get_endpoint(plugin_info['plugin_id'],
                                                                plugin_info.get('version'),
                                                                domain_id,
                                                                plugin_info.get('upgrade_mode', 'AUTO'))

        plugin_response = collector_plugin_mgr.init_plugin(endpoint, plugin_info.get('options', {}))

        if updated_version:
            plugin_info['version'] = updated_version

        plugin_info['metadata'] = plugin_response.get('metadata', {})
        plugin_info_params = {'plugin_info': plugin_info}

        collector_vo = collector_mgr.update_collector_by_vo(collector_vo, plugin_info_params)
        self.create_collector_rules_by_metadata(plugin_info['metadata'], collector_vo.collector_id, params['domain_id'])

        return collector_vo

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_id', 'domain_id'])
    def update(self, params):
        """
        Args:
            params (dict): {
                'collector_id': 'str',
                'name': 'str',
                'priority': 'int',
                'tags': 'dict',
                'schedule': 'dict',
                'secret_filter': 'dict',
                'domain_id': 'str'
            }

        Returns:
            collector_vo (object)
        """
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        collector_vo = collector_mgr.get_collector(params['collector_id'], params['domain_id'])

        if 'tags' in params:
            params['tags'] = self._convert_tags(params.get('tags'))

        if 'secret_filter' in params:
            self.validate_secret_filter(params['secret_filter'], params['domain_id'])

        return collector_mgr.update_collector_by_vo(collector_vo, params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_id', 'domain_id'])
    def delete(self, params):
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        state_mgr: CollectionStateManager = self.locator.get_manager(CollectionStateManager)

        collector_mgr.delete_collector(params['collector_id'], params['domain_id'])
        state_mgr.delete_collection_state_by_collector_id(params['collector_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_id', 'domain_id'])
    def get(self, params):
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        collector_id = params['collector_id']
        domain_id = params['domain_id']
        only = params.get('only')
        return collector_mgr.get_collector(collector_id, domain_id, only)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @append_query_filter(['collector_id', 'name', 'state', 'priority', 'plugin_id', 'domain_id'])
    @append_keyword_filter(_KEYWORD_FILTER)
    def list(self, params):
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        return collector_mgr.list_collectors(params.get('query', {}))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(_KEYWORD_FILTER)
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list) : 'list of statistics data'

        """
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        return collector_mgr.stat_collectors(params.get('query', {}))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_id', 'domain_id'])
    def collect(self, params):
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        plugin_mgr: PluginManager = self.locator.get_manager(PluginManager)
        job_mgr: JobManager = self.locator.get_manager(JobManager)
        job_task_mgr: JobTaskManager = self.locator.get_manager(JobTaskManager)

        collector_id = params['collector_id']
        domain_id = params['domain_id']
        collector_vo = collector_mgr.get_collector(collector_id, domain_id)
        collector_dict = collector_vo.to_dict()

        plugin_info = collector_dict['plugin_info']
        secret_filter = collector_dict.get('secret_filter', {})
        plugin_id = plugin_info['plugin_id']
        version = plugin_info.get('version')
        upgrade_mode = plugin_info.get('upgrade_mode', 'AUTO')

        endpoint, updated_version = plugin_mgr.get_endpoint(plugin_id, version, domain_id, upgrade_mode)

        if updated_version and version != updated_version:
            _LOGGER.debug(f'[collect] upgrade plugin version: {version} -> {updated_version}')
            collector_vo = self._update_collector_plugin(endpoint, updated_version, plugin_info, collector_vo, domain_id)

        tasks = self.get_tasks(params, endpoint, collector_vo.provider, plugin_info, secret_filter, domain_id)
        projects = self.list_projects_from_tasks(tasks)
        params.update({'plugin_id': plugin_id, 'total_tasks': len(tasks), 'remained_tasks': len(tasks)})

        duplicated_job_vos = job_mgr.list_duplicate_jobs(collector_id, params.get('secret_id'), domain_id)
        for job_vo in duplicated_job_vos:
            job_mgr.make_canceled_by_vo(job_vo)

        # JOB: IN-PROGRESS
        job_vo = job_mgr.create_job(collector_vo, params)

        # job_mgr.make_inprogress_by_vo(job_vo)

        if tasks:
            for task in tasks:
                task_options = task['task_options']
                task.update({'collector_id': collector_id, 'job_id': job_vo.job_id})

                try:
                    # JOB: CREATE TASK JOB
                    job_task_vo = job_task_mgr.create_job_task(job_vo, domain_id, task_options)
                    task.update({'job_task_id': job_task_vo.job_task_id})
                    job_task_mgr.push_job_task(task)

                except Exception as e:
                    job_mgr.add_error(job_vo.job_id, domain_id, 'ERROR_COLLECTOR_COLLECTING', e, {'task_options': task_options})
                    _LOGGER.error(f'[collect] collecting failed: task_options={task_options}: {e}')
        else:
            # JOB: SUCCESS (No tasks)
            job_mgr.make_success_by_vo(job_vo)
            return job_vo

        collector_mgr.update_last_collected_time(collector_vo)
        return job_mgr.update_job_by_vo({'projects': projects}, job_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_id', 'domain_id'])
    def update_plugin(self, params):
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        plugin_manager: PluginManager = self.locator.get_manager(PluginManager)

        collector_id = params['collector_id']
        domain_id = params['domain_id']

        collector_vo = collector_mgr.get_collector(collector_id, domain_id)
        plugin_info = collector_vo.plugin_info.to_dict()

        if version := params.get('version'):
            plugin_info['version'] = version

        if options := params.get('options'):
            plugin_info['options'] = options

        if upgrade_mode := params.get('upgrade_mode'):
            plugin_info['upgrade_mode'] = upgrade_mode

        endpoint, updated_version = plugin_manager.get_endpoint(plugin_info['plugin_id'],
                                                                plugin_info.get('version'),
                                                                domain_id,
                                                                plugin_info.get('upgrade_mode', 'AUTO'))

        collector_vo = self._update_collector_plugin(endpoint, updated_version, plugin_info, collector_vo, domain_id)
        return collector_vo

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['collector_id', 'domain_id'])
    def verify_plugin(self, params):
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        collector_plugin_mgr: CollectorPluginManager = self.locator.get_manager(CollectorPluginManager)
        plugin_manager: PluginManager = self.locator.get_manager(PluginManager)
        secret_manager: SecretManager = self.locator.get_manager(SecretManager)

        collector_id = params['collector_id']
        domain_id = params['domain_id']

        collector_vo = collector_mgr.get_collector(collector_id, domain_id)
        plugin_info = collector_vo.plugin_info.to_dict()

        endpoint, updated_version = plugin_manager.get_endpoint(plugin_info['plugin_id'],
                                                                plugin_info.get('version'),
                                                                domain_id,
                                                                plugin_info.get('upgrade_mode', 'AUTO'))

        secret_ids = self.list_secret_from_secret_filter(plugin_info.get('secret_filter', {}),
                                                         params.get('secret_id'),
                                                         collector_vo.provider,
                                                         domain_id)

        if secret_ids:
            secret_data_info = secret_manager.get_secret_data(secret_ids[0], domain_id)
            secret_data = secret_data_info.get('data', {})
            collector_plugin_mgr.verify_plugin(endpoint, plugin_info.get('options', {}), secret_data)

    def get_tasks(self, params, endpoint, collector_provider, plugin_info, secret_filter, domain_id):
        secret_mgr: SecretManager = self.locator.get_manager(SecretManager)
        collector_plugin_mgr: CollectorPluginManager = self.locator.get_manager(CollectorPluginManager)

        tasks = []
        secret_ids = self.list_secret_from_secret_filter(secret_filter, params.get('secret_id'), collector_provider, domain_id)

        for secret_id in secret_ids:
            secret_info = secret_mgr.get_secret(secret_id, domain_id)
            secret_data = secret_mgr.get_secret_data(secret_id, domain_id)

            _task = {
                'plugin_info': plugin_info,
                'secret_info': secret_info,
                'secret_data': secret_data,
                'domain_id': domain_id
            }

            try:
                response = collector_plugin_mgr.get_tasks(endpoint, secret_data, plugin_info.get('options', {}))
                _LOGGER.debug(f'[get_tasks] response: {response}')

                for task_options in response.get('tasks', []):
                    _task_dict = copy.deepcopy(_task)
                    _task_dict.update(task_options)
                    tasks.append(_task_dict)

            except Exception as e:
                # _LOGGER.debug(f'[get_tasks] Error to get tasks from plugin. set task from secret')
                _task.update({'task_options': None})
                tasks.append(_task)

        return tasks

    def validate_secret_filter(self, secret_filter, domain_id):
        if 'secrets' in secret_filter:
            _query = {'filter': [{'k': 'secret_id', 'v': secret_filter['secrets'], 'o': 'in'}]}
            secret_mgr: SecretManager = self.locator.get_manager(SecretManager)
            response = secret_mgr.list_secrets(_query, domain_id)
            if response.get('total_count', 0) != len(secret_filter['secrets']):
                raise ERROR_INVALID_PARAMETER(key='secret_filter.secrets', reason='Secrets not found')
        if 'service_accounts' in secret_filter:
            _query = {'filter': [{'k': 'service_account_id', 'v': secret_filter['service_accounts'], 'o': 'in'}]}
            identity_mgr: IdentityManager = self.locator.get_manager(IdentityManager)
            response = identity_mgr.list_service_accounts(_query, domain_id)
            if response.get('total_count', 0) != len(secret_filter['service_accounts']):
                raise ERROR_INVALID_PARAMETER(key='secret_filter.service_accounts', reason='Service accounts not found')
        if 'schemas' in secret_filter:
            _query = {'filter': [{'k': 'name', 'v': secret_filter['schemas'], 'o': 'in'}]}
            repo_mgr: RepositoryManager = self.locator.get_manager(RepositoryManager)
            response = repo_mgr.list_schemas(_query, domain_id)
            if response.get('total_count', 0) != len(secret_filter['schemas']):
                raise ERROR_INVALID_PARAMETER(key='secret_filter.schema', reason='Schema not found')

    def _update_collector_plugin(self, endpoint, updated_version, plugin_info, collector_vo, domain_id):
        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        collector_plugin_mgr: CollectorPluginManager = self.locator.get_manager(CollectorPluginManager)
        plugin_response = collector_plugin_mgr.init_plugin(endpoint, plugin_info.get('options', {}))

        if updated_version:
            plugin_info['version'] = updated_version

        plugin_info['metadata'] = plugin_response.get('metadata', {})

        params = {'plugin_info': plugin_info}
        collector_vo = collector_mgr.update_collector_by_vo(collector_vo, params)

        self.delete_collector_rules(collector_vo.collector_id, collector_vo.domain_id),
        self.create_collector_rules_by_metadata(plugin_info['metadata'], collector_vo.collector_id, domain_id)

        return collector_vo

    def list_secret_from_secret_filter(self, secret_filter, secret_id, collector_provider, domain_id):
        secret_manager: SecretManager = self.locator.get_manager(SecretManager)

        _filter = self._set_secret_filter(secret_filter, secret_id, collector_provider)
        query = {'filter': _filter} if _filter else {}
        response = secret_manager.list_secrets(query, domain_id)

        return [secret_info.get('secret_id') for secret_info in response.get('results', [])]

    @check_required(['schedule'])
    def scheduled_collectors(self, params):
        """ Search all collectors in this schedule
        This is global search out-of domain

        Args:
            params(dict): {
                schedule(dict): {
                  'hours': list,
                  'minutes': list
                }
                domain_id: optional
            }
        ex) {'hour': 3}

        Returns: collectors_info
        """

        collector_mgr: CollectorManager = self.locator.get_manager(CollectorManager)
        filter_query = [{'k': 'schedule.state', 'v': 'ENABLED', 'o': 'eq'}]

        if 'domain_id' in params:
            filter_query.append({'k': 'domain_id', 'v': params['domain_id'], 'o': 'eq'})

        schedule = params.get('schedule', {})
        if 'hour' in schedule:
            filter_query.append({'k': 'schedule.hours', 'v': schedule['hour'], 'o': 'contain'})

        return collector_mgr.list_collectors({'filter': filter_query})

    def _get_plugin_from_repository(self, plugin_info, domain_id):
        repo_mgr: RepositoryManager = self.locator.get_manager(RepositoryManager)
        return repo_mgr.get_plugin(plugin_info['plugin_id'], domain_id)

    def create_collector_rules_by_metadata(self, plugin_metadata, domain_id, collector_id):
        collector_rule_mgr: CollectorRuleManager = self.locator.get_manager(CollectorRuleManager)
        collector_rules = plugin_metadata.get('collector_rules', [])

        for collector_rule_params in collector_rules:
            collector_rule_params.update({
                'domain_id': domain_id,
                'collector_id': collector_id,
                'rule_type': 'MANAGED'
            })

            collector_rule_mgr.create_collector_rule(collector_rule_params)

    def delete_collector_rules(self, collector_id, domain_id):
        collector_rule_mgr: CollectorRuleManager = self.locator.get_manager('CollectorRuleManager')
        old_collector_rule_vos = collector_rule_mgr.filter_collector_rules(collector_id=collector_id,
                                                                           rule_type='MANAGED',
                                                                           domain_id=domain_id)
        old_collector_rule_vos.delete()

    def _set_secret_info(self, secret_id, domain_id):
        secret_mgr: SecretManager = self.locator.get_manager('SecretManager')
        secret = secret_mgr.get_secret(secret_id, domain_id)
        secret_info = {'secret_id': secret_id}

        if provider := secret.get('provider'):
            secret_info.update({'provider': provider})
        if service_account_id := secret.get('service_account_id'):
            secret_info.update({'service_account_id': service_account_id})

        if project_id := secret.get('project_id'):
            secret_info.update({'project_id': project_id})

        return secret_info

    @staticmethod
    def _set_secret_filter(secret_filter, secret_id, collector_provider):
        _filter = []

        if secret_id:
            _filter.append({'k': 'secret_id', 'v': secret_id, 'o': 'eq'})

        if collector_provider:
            _filter.append({'k': 'provider', 'v': collector_provider, 'o': 'eq'})

        if secret_filter and secret_filter.get('state') == 'ENABLED':
            if 'secrets' in secret_filter and secret_filter['secrets']:
                _filter.append({'k': 'secret_id', 'v': secret_filter['secrets'], 'o': 'in'})
            if 'service_accounts' in secret_filter and secret_filter['service_accounts']:
                _filter.append({'k': 'service_account_id', 'v': secret_filter['service_accounts'], 'o': 'in'})
            if 'schemas' in secret_filter and secret_filter['schemas']:
                _filter.append({'k': 'schema', 'v': secret_filter['schemas'], 'o': 'in'})
            if 'exclude_secrets' in secret_filter and secret_filter['exclude_secrets']:
                _filter.append({'k': 'secret_id', 'v': secret_filter['exclude_secrets'], 'o': 'not_in'})
            if 'exclude_service_accounts' in secret_filter and secret_filter['exclude_service_accounts']:
                _filter.append({'k': 'service_account_id', 'v': secret_filter['exclude_service_accounts'], 'o': 'not_in'})
            if 'exclude_schemas' in secret_filter and secret_filter['exclude_schemas']:
                _filter.append({'k': 'exclude_schemas', 'v': secret_filter['exclude_schemas'], 'o': 'not_in'})

        return _filter

    @staticmethod
    def _get_plugin_providers(provider, plugin_info):
        supported_providers = plugin_info.get('capability', {}).get('supported_providers', [])

        if supported_providers:
            # Multi providers
            if provider in supported_providers:
                return provider
            else:
                raise ERROR_INVALID_PARAMETER(key='provider', reason=f'Not supported provider: {provider}')
        else:
            # Single provider
            return provider if provider else plugin_info.get('provider')

    @staticmethod
    def _convert_tags(tags):
        if isinstance(tags, list):
            return utils.tags_to_dict(tags)
        else:
            return tags

    @staticmethod
    def list_projects_from_tasks(tasks):
        projects = []
        for task in tasks:
            if project_id := task['secret_info'].get('project_id'):
                projects.append(project_id)

        return list(set(projects))



import logging
import copy
from datetime import datetime

from spaceone.core.manager import BaseManager
from spaceone.inventory.model.cloud_service_model import CloudService
from spaceone.inventory.lib.resource_manager import ResourceManager
from spaceone.inventory.manager.collection_state_manager import CollectionStateManager

_LOGGER = logging.getLogger(__name__)

MERGE_KEYS = [
    'name',
    'ip_addresses',
    'account',
    'instance_type',
    'instance_size',
    'reference'
    'region_code',
    'ref_region',
    'project_id',
    'data'
]


class CloudServiceManager(BaseManager, ResourceManager):
    resource_keys = ['cloud_service_id']
    query_method = 'list_cloud_services'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cloud_svc_model: CloudService = self.locator.get_model('CloudService')

    def create_cloud_service(self, params):
        def _rollback(cloud_svc_vo: CloudService):
            _LOGGER.info(
                f'[ROLLBACK] Delete Cloud Service : {cloud_svc_vo.provider} ({cloud_svc_vo.cloud_service_type})')
            cloud_svc_vo.terminate()

        cloud_svc_vo: CloudService = self.cloud_svc_model.create(params)
        self.transaction.add_rollback(_rollback, cloud_svc_vo)

        return cloud_svc_vo

    def update_cloud_service(self, params):
        return self.update_cloud_service_by_vo(params,
                                               self.get_cloud_service(params['cloud_service_id'], params['domain_id']))

    def update_cloud_service_by_vo(self, params, cloud_svc_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[ROLLBACK] Revert Data : {old_data.get("cloud_service_id")}')
            cloud_svc_vo.update(old_data)

        self.transaction.add_rollback(_rollback, cloud_svc_vo.to_dict())
        cloud_svc_vo: CloudService = cloud_svc_vo.update(params)

        return cloud_svc_vo

    def delete_cloud_service(self, cloud_service_id, domain_id):
        cloud_svc_vo = self.get_cloud_service(cloud_service_id, domain_id)
        cloud_svc_vo.delete()

    @staticmethod
    def delete_cloud_service_by_vo(cloud_svc_vo: CloudService):
        cloud_svc_vo.delete()

    def terminate_cloud_service(self, cloud_service_id, domain_id):
        cloud_svc_vo: CloudService = self.get_cloud_service(cloud_service_id, domain_id)
        cloud_svc_vo.terminate()

    def get_cloud_service(self, cloud_service_id, domain_id, user_projects=None, only=None):
        kwargs = {
            'cloud_service_id': cloud_service_id,
            'domain_id': domain_id
        }

        if user_projects:
            kwargs['project_id'] = user_projects

        if only:
            kwargs['only'] = only

        return self.cloud_svc_model.get(**kwargs)

    def list_cloud_services(self, query, target=None):
        # Append Query for DELETED filter (Temporary Logic)
        query = self._append_state_query(query)
        return self.cloud_svc_model.query(**query, target=target)

    def analyze_cloud_services(self, query):
        # Append Query for DELETED filter (Temporary Logic)
        query = self._append_state_query(query)
        return self.cloud_svc_model.analyze(**query)

    def stat_cloud_services(self, query):
        # Append Query for DELETED filter (Temporary Logic)
        query = self._append_state_query(query)
        return self.cloud_svc_model.stat(**query)

    @staticmethod
    def merge_data(new_data, old_data):
        for key in MERGE_KEYS:
            if key in new_data:
                new_value = new_data[key]
                old_value = old_data.get(key)
                if key in ['data', 'tags']:
                    is_changed = False
                    for sub_key, sub_value in new_value.items():
                        if sub_value != old_value.get(sub_key):
                            is_changed = True
                            break

                    if is_changed:
                        merged_value = copy.deepcopy(old_value)
                        merged_value.update(new_value)
                        new_data[key] = merged_value
                    else:
                        del new_data[key]
                else:
                    if new_value == old_value:
                        del new_data[key]

        return new_data

    def find_resources(self, query):
        query['only'] = ['cloud_service_id']

        resources = []
        cloud_svc_vos, total_count = self.list_cloud_services(query, target='SECONDARY_PREFERRED')

        for cloud_svc_vo in cloud_svc_vos:
            resources.append({
                'cloud_service_id': cloud_svc_vo.cloud_service_id
            })

        return resources, total_count

    def delete_resources(self, query):
        query['only'] = self.resource_keys

        vos, total_count = self.list_cloud_services(query)

        resource_ids = []
        for vo in vos:
            resource_ids.append(vo.cloud_service_id)

        vos.update({
            'state': 'DELETED',
            'deleted_at': datetime.utcnow()
        })

        state_mgr: CollectionStateManager = self.locator.get_manager('CollectionStateManager')
        state_mgr.delete_collection_state_by_resource_ids(resource_ids)

        return total_count

    @staticmethod
    def _append_state_query(query):
        state_default_filter = {
            'key': 'state',
            'value': 'ACTIVE',
            'operator': 'eq'
        }

        show_deleted_resource = False
        for condition in query.get('filter', []):
            key = condition.get('k', condition.get('key'))
            value = condition.get('v', condition.get('value'))
            operator = condition.get('o', condition.get('operator'))

            if key == 'state':
                if operator == 'eq' and value == 'DELETED':
                    show_deleted_resource = True
                elif operator in ['in', 'contain_in'] and 'DELETED' in value:
                    show_deleted_resource = True

        if not show_deleted_resource:
            query['filter'] = query.get('filter', [])
            query['filter'].append(state_default_filter)

        return query



import logging
import json
import time
from spaceone.core import config, cache
from spaceone.core import queue
# from spaceone.core.transaction import LOCAL_STORAGE
from spaceone.core.manager import BaseManager
from spaceone.inventory.manager.job_manager import JobManager
from spaceone.inventory.manager.job_task_manager import JobTaskManager
from spaceone.inventory.manager.plugin_manager import PluginManager
from spaceone.inventory.manager.cleanup_manager import CleanupManager
from spaceone.inventory.manager.collector_plugin_manager import CollectorPluginManager
from spaceone.inventory.error import *
from spaceone.inventory.lib import rule_matcher
from spaceone.inventory.conf.collector_conf import *

_LOGGER = logging.getLogger(__name__)


class CollectingManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.job_mgr: JobManager = self.locator.get_manager(JobManager)
        self.job_task_mgr: JobTaskManager = self.locator.get_manager(JobTaskManager)
        self.db_queue = DB_QUEUE_NAME

    def collecting_resources(self, params):
        """ Execute collecting task to get resources from plugin
        Args:
            params(dict): {
                collector_id(str)
                job_id(str)
                job_task_id(str)
                domain_id(str)
                plugin_info(dict)
                task_options(dict)
                secret_info(dict)
                secret_data(dict)
            }
        """
        plugin_manager: PluginManager = self.locator.get_manager(PluginManager)
        collector_plugin_mgr: CollectorPluginManager = self.locator.get_manager(CollectorPluginManager)

        collector_id = params['collector_id']
        job_id = params['job_id']
        job_task_id = params['job_task_id']
        domain_id = params['domain_id']
        task_options = params['task_options']

        secret_info = params['secret_info']
        secret_id = secret_info['secret_id']
        secret_data = params['secret_data']
        plugin_info = params['plugin_info']

        _LOGGER.debug(f'[collecting_resources] Job Task ID: {job_task_id}')

        if self.job_mgr.check_cancel(job_id, domain_id):
            self.job_task_mgr.add_error(job_task_id, domain_id, 'ERROR_COLLECT_CANCELED', 'The job has been canceled.')
            self.job_task_mgr.make_failure(job_task_id, domain_id)
            self.job_mgr.decrease_remained_tasks(job_id, domain_id)
            self.job_mgr.increase_failure_tasks(job_id, domain_id)
            raise ERROR_COLLECT_CANCELED(job_id=job_id)

        collect_filter = {}

        try:
            # JOB TASK: IN_PROGRESS
            self._update_job_task(job_task_id, 'IN_PROGRESS', domain_id, secret_info=secret_info)
        except Exception as e:
            _LOGGER.error(f'[collecting_resources] fail to update job_task: {e}')

        try:
            # EXECUTE PLUGIN COLLECTION
            endpoint, updated_version = plugin_manager.get_endpoint(plugin_info['plugin_id'], plugin_info.get('version'), domain_id, plugin_info.get('upgrade_mode', 'AUTO'))
            results = collector_plugin_mgr.collect(endpoint, plugin_info['options'], secret_data.get('data', {}), collect_filter, task_options)     # task_options = None
            # DELETE secret_data in params for Secure
            del params['secret_data']
        except Exception as e:
            _LOGGER.error(f'[collecting_resources - ERROR {job_id}/{job_task_id}] {e}', exc_info=True)
            self.job_task_mgr.add_error(job_task_id, domain_id, 'ERROR_COLLECTOR_COLLECTING', e)
            self.job_task_mgr.make_failure(job_task_id, domain_id)
            self.job_mgr.decrease_remained_tasks(job_id, domain_id)
            self.job_mgr.increase_failure_tasks(job_id, domain_id)
            raise ERROR_COLLECTOR_COLLECTING(plugin_info=plugin_info, filters=collect_filter)

        JOB_TASK_STATE = 'SUCCESS'
        collecting_count_info = {}

        try:
            collecting_count_info = self._check_collecting_results(results, params)

            if collecting_count_info['failure_count'] > 0:
                JOB_TASK_STATE = 'FAILURE'

        except Exception as e:
            _LOGGER.error(f'[collecting_resources] {e}', exc_info=True)
            self.job_task_mgr.add_error(job_task_id, domain_id, 'ERROR_COLLECTOR_COLLECTING', e)
            self.job_task_mgr.make_failure(job_task_id, domain_id)
            JOB_TASK_STATE = 'FAILURE'

        finally:
            cleanup_mode = self._check_garbage_collection_mode(plugin_info)

            if cleanup_mode and JOB_TASK_STATE == 'SUCCESS':
                disconnected_count, deleted_count = self._update_disconnected_and_deleted_count(collector_id, secret_id, job_task_id, domain_id)
                collecting_count_info.update({'disconnected_count': disconnected_count, 'deleted_count': deleted_count})
                _LOGGER.debug(f'[collecting_resources] {job_task_id} | disconnected: {disconnected_count}, deleted: {deleted_count}')
            else:
                _LOGGER.debug(f'[collecting_resources] skip garbage_collection, {cleanup_mode}, {JOB_TASK_STATE}')

            _LOGGER.debug(f'[collecting_resources] {job_task_id} | collecting_count_info: {collecting_count_info}')
            self._update_job_task(job_task_id, JOB_TASK_STATE, domain_id, secret_info=secret_info, collecting_count_info=collecting_count_info)
            self.job_mgr.decrease_remained_tasks(job_id, domain_id)

            if JOB_TASK_STATE == 'SUCCESS':
                self.job_mgr.increase_success_tasks(job_id, domain_id)
            elif JOB_TASK_STATE == 'FAILURE':
                self.job_mgr.increase_failure_tasks(job_id, domain_id)

            # debug code for memory leak
            # local_storage = LOCAL_STORAGE.__dict__
            # _LOGGER.info(
            #     f'[collecting_resources] / number of items in local storage: {len(local_storage)} / items => {local_storage}')
        return True

    def _update_disconnected_and_deleted_count(self, collector_id, secret_id, job_task_id, domain_id):
        try:
            cleanup_mgr: CleanupManager = self.locator.get_manager(CleanupManager)
            return cleanup_mgr.update_disconnected_and_deleted_count(collector_id, secret_id, job_task_id, domain_id)
        except Exception as e:
            _LOGGER.error(f'[_update_collection_state] failed: {e}')
            return 0, 0

    def _check_collecting_results(self, results, params):
        """
        Args:
            params(dict): {
                collector_id(str)
                job_id(str)
                job_task_id(str)
                domain_id(str)
                plugin_info(dict)
                task_options(dict)
                secret_info(dict)
            }
        """
        created_count = 0
        updated_count = 0
        failure_count = 0
        total_count = 0

        self._set_transaction_meta(params)

        for res in results:
            total_count += 1
            try:
                res_state = self.check_resource_state(res, params)

                if res_state == NOT_COUNT:
                    pass
                elif res_state == CREATED:
                    created_count += 1
                elif res_state == UPDATED:
                    updated_count += 1
                else:
                    failure_count += 1

            except Exception as e:
                _LOGGER.error(f'[_process_results] failed single result {e}')
                failure_count += 1

        return {
            'total_count': total_count,
            'created_count': created_count,
            'updated_count': updated_count,
            'failure_count': failure_count
        }

    def check_resource_state(self, resource, params):
        """
        Args:
            resource (message_dict): resource from collector
            params(dict): {
                collector_id(str)
                job_id(str)
                job_task_id(str)
                domain_id(str)
                plugin_info(dict)
                task_options(dict)
                secret_info(dict)
            }
        Returns:
            0: NOT_COUNT (for example, cloud_service_type)
            1: CREATED
            2: UPDATED
            3: ERROR
        """
        job_task_id = params['job_task_id']
        domain_id = params['domain_id']
        resource_type = resource.get('resource_type')
        state = resource.get('state', 'None')
        data = resource.get('resource', {})

        if update_mode := resource.get('options', {}).get('update_mode'):   # MERGE | REPLACE
            self.transaction.set_meta('update_mode', update_mode)

        resource_service, resource_manager = self._get_resource_map(resource_type)
        data['domain_id'] = domain_id
        response = ERROR

        if resource_type == "inventory.ErrorResource" and state == "FAILURE":
            err_message = resource.get('message', 'No message from plugin')
            _LOGGER.error(f'[_process_single_result] Error resource: {err_message}')
            self.job_task_mgr.add_error(job_task_id, domain_id, "ERROR_PLUGIN", err_message, data)
            return ERROR

        match_rules = resource.get('match_rules')

        if not match_rules:
            _msg = 'No match rule'
            _LOGGER.error(f'[_process_single_result] {_msg}')
            self.job_task_mgr.add_error(job_task_id, domain_id, "ERROR_MATCH_RULE", f"{_msg}: {resource}", {'resource_type': resource_type})
            return ERROR

        try:
            res_info, total_count = self._query_with_match_rules(data, match_rules, domain_id, resource_manager)
        except ERROR_TOO_MANY_MATCH as e:
            _LOGGER.error(f'[_process_single_result] Too many match')
            self.job_task_mgr.add_error(job_task_id, domain_id, e.error_code, e.message, {'resource_type': resource_type})
            return ERROR
        except Exception as e:
            _LOGGER.error(f'[_process_single_result] failed to match: {e}')
            self.job_task_mgr.add_error(job_task_id, domain_id, "ERROR_UNKNOWN", "Match Query failed, may be DB problem", {'resource_type': resource_type})
            return ERROR

        try:
            # CREATE
            if total_count == 0 and update_mode is None:
                resource_service.create_resource(data)
                response = CREATED
            # UPDATE
            elif total_count == 1:
                data.update(res_info[0])
                resource_service.update_resource(data)
                response = UPDATED
            elif total_count > 1:
                _LOGGER.error(f'[_process_single_result] will not reach here!')
                response = ERROR
        except ERROR_BASE as e:
            _LOGGER.error(f'[_process_single_result] ERROR: {data}')
            additional = self._set_error_addition_info(resource_type, total_count, data)
            self.job_task_mgr.add_error(job_task_id, domain_id, e.error_code, e.message, additional)
            response = ERROR
        except Exception as e:
            # TODO: create error message
            _LOGGER.debug(f'[_process_single_result] service error: {resource_service}, {e}')
            response = ERROR

        finally:
            if response in [CREATED, UPDATED]:
                if resource_type in ['inventory.CloudServiceType', 'inventory.Region']:
                    response = NOT_COUNT
            return response

    def _set_transaction_meta(self, params):
        secret_info = params['secret_info']

        self.transaction.set_meta('job_id', params['job_id'])
        self.transaction.set_meta('job_task_id', params['job_task_id'])
        self.transaction.set_meta('collector_id', params['collector_id'])
        self.transaction.set_meta('secret.secret_id', secret_info['secret_id'])
        self.transaction.set_meta('disable_info_log', 'true')

        if plugin_id := params['plugin_info'].get('plugin_id'):
            self.transaction.set_meta('plugin_id', plugin_id)
        if 'provider' in secret_info:
            self.transaction.set_meta('secret.provider', secret_info['provider'])

        if 'project_id' in secret_info:
            self.transaction.set_meta('secret.project_id', secret_info['project_id'])
        if 'service_account_id' in secret_info:
            self.transaction.set_meta('secret.service_account_id', secret_info['service_account_id'])

    def _get_resource_map(self, resource_type):
        if resource_type not in RESOURCE_MAP:
            raise ERROR_UNSUPPORTED_RESOURCE_TYPE(resource_type=resource_type)

        svc = self.locator.get_service(RESOURCE_MAP[resource_type][0])
        mgr = self.locator.get_manager(RESOURCE_MAP[resource_type][1])
        return svc, mgr

    def _update_job_task(self, job_task_id, state, domain_id, secret_info=None, collecting_count_info=None):
        state_map = {
            'IN_PROGRESS': self.job_task_mgr.make_inprogress,
            'SUCCESS': self.job_task_mgr.make_success,
            'FAILURE': self.job_task_mgr.make_failure,
            'CANCELED': self.job_task_mgr.make_canceled
        }

        if secret_info:
            secret_info = {
                'secret_id': secret_info['secret_id'],
                'provider': secret_info.get('provider'),
                'service_account_id': secret_info.get('service_account_id'),
                'project_id': secret_info.get('project_id')
            }

        if state in state_map:
            state_map[state](job_task_id, domain_id, secret_info, collecting_count_info)
        else:
            _LOGGER.error(f'[_update_job_task] undefined state: {state}')
            self.job_task_mgr.make_failure(job_task_id, domain_id, secret_info, collecting_count_info)

    @staticmethod
    def _set_error_addition_info(resource_type, total_count, data):
        additional = {'resource_type': resource_type}

        if resource_type == 'inventory.CloudService':
            additional.update({
                'cloud_service_group': data.get('cloud_service_group'),
                'cloud_service_type': data.get('cloud_service_type'),
                'provider': data.get('provider'),
            })

        if total_count == 1:
            if resource_type == 'inventory.CloudService':
                additional['resource_id'] = data.get('cloud_service_id')
            elif resource_type == 'inventory.CloudServiceType':
                additional['resource_id'] = data.get('cloud_service_type_id')
            elif resource_type == 'inventory.Region':
                additional['resource_id'] = data.get('region_id')

        return additional

    @staticmethod
    def _delete_job_task_stat_cache(job_id, job_task_id, domain_id):
        try:
            for state in ['CREATED', 'UPDATED', 'FAILURE']:
                cache.delete(f'job_task_stat:{domain_id}:{job_id}:{job_task_id}:{state}')
        except Exception as e:
            _LOGGER.error(f'[_delete_job_task_stat_cache] {e}')

    @staticmethod
    def _collector_filter_from_cache(collect_filter, collector_id, secret_id):
        key = f'collector-filter:{collector_id}:{secret_id}'

        if value := cache.get(key):
            _LOGGER.debug(f'[collecting_resources] cache -> {key}: {value}')
            collect_filter.update(value)

        return collect_filter

    def watchdog_job_task_stat(self, param):
        _LOGGER.debug(f'[watchdog_job_task_stat] WatchDog Start: {param["job_task_id"]}')

        domain_id = param['domain_id']
        job_id = param['job_id']
        job_task_id = param['job_task_id']

        time.sleep(WATCHDOG_WAITING_TIME)

        stat_result = {
            'total_count': param['total_count'],
            'created_count': 0,
            'updated_count': 0,
            'failure_count': 0
        }

        try:
            key_created = f'job_task_stat:{domain_id}:{job_id}:{job_task_id}:CREATED'
            stat_result['created_count'] = cache.get(key_created)

            key_updated = f'job_task_stat:{domain_id}:{job_id}:{job_task_id}:UPDATED'
            stat_result['updated_count'] = cache.get(key_updated)

            key_failure = f'job_task_stat:{domain_id}:{job_id}:{job_task_id}:FAILURE'
            stat_result['failure_count'] = cache.get(key_failure)
        except Exception as e:
            _LOGGER.error(f'[watchdog_job_task_stat] Get Count from Cache: {e}')
            pass

        self._delete_job_task_stat_cache((job_id, job_task_id, domain_id))

        try:
            if stat_result['failure_count'] > 0:
                JOB_TASK_STATE = 'FAILURE'
            else:
                JOB_TASK_STATE = 'SUCCESS'
            self._update_job_task(job_task_id, JOB_TASK_STATE, domain_id, collecting_count_info=stat_result)
        except Exception as e:
            pass

        finally:
            self.job_mgr.decrease_remained_tasks(job_id, domain_id)

    @staticmethod
    def _check_garbage_collection_mode(plugin_info):
        return True if 'garbage_collection' in plugin_info.get('metadata', {}).get('supported_features', []) else False

    @staticmethod
    def _query_with_match_rules(resource, match_rules, domain_id, resource_manager):
        """ match resource based on match_rules

        Args:
            resource: ResourceInfo(Json) from collector plugin
            match_rules:
                ex) {1:['data.vm.vm_id'], 2:['zone_id', 'data.ip_addresses']}

        Return:
            match_resource : resource_id for resource update (ex. {'server_id': 'server-xxxxxx'})
            total_count : total count of matched resources
        """
        match_resource = None
        total_count = 0

        match_rules = rule_matcher.dict_key_int_parser(match_rules)

        for order in sorted(match_rules.keys()):
            query = rule_matcher.make_query(order, match_rules, resource, domain_id)
            match_resource, total_count = resource_manager.find_resources(query)

            if total_count > 1:
                data = resource['data'] if 'data' in resource else resource
                raise ERROR_TOO_MANY_MATCH(match_key=match_rules[order], resources=match_resource, more=data)
            elif total_count == 1 and match_resource:
                return match_resource, total_count

        return match_resource, total_count

    """
    Deprecated
    """
    def _create_db_update_task_watchdog(self, total_count, job_id, job_task_id, domain_id):
        try:
            # PUSH QUEUE
            param = {'job_id': job_id, 'job_task_id': job_task_id, 'domain_id': domain_id, 'total_count': total_count}
            task = {'method': 'watchdog_job_task_stat', 'res': {}, 'param': param, 'meta': self.transaction.meta}
            json_task = json.dumps(task)
            queue.put(self.db_queue, json_task)
            return True
        except Exception as e:
            _LOGGER.error(f'[_create_db_update_task_watchdog] {e}')
            return False

    """
    Deprecated
    """
    def _create_db_update_task(self, res, param):
        try:
            _LOGGER.debug(f'[_create_db_update_task] Push Update Task Queue')
            task = {'method': 'check_resource_state', 'res': res, 'param': param, 'meta': self.transaction.meta}
            json_task = json.dumps(task)
            queue.put(self.db_queue, json_task)
            return True
        except Exception as e:
            _LOGGER.error(f'[_create_db_update_task] {e}')
            return False

    """
    Deprecated
    """
    @staticmethod
    def _update_job_task_stat_to_cache(job_id, job_task_id, kind, domain_id):
        if kind in [CREATED, UPDATED]:
            key = f'job_task_stat:{domain_id}:{job_id}:{job_task_id}:{kind}'
        else:
            key = f'job_task_stat:{domain_id}:{job_id}:{job_task_id}:FAILURE'

        cache.increment(key)

    """
    Deprecated
    """
    @staticmethod
    def _create_job_task_stat_cache(job_id, job_task_id, domain_id):
        try:
            for state in ['CREATED', 'UPDATED', 'FAILURE']:
                cache.set(f'job_task_stat:{domain_id}:{job_id}:{job_task_id}:{state}', 0, expire=JOB_TASK_STAT_EXPIRE_TIME)
        except Exception as e:
            _LOGGER.error(f'[_create_job_task_stat_cache] {e}')



from spaceone.inventory.info.common_info import *
from spaceone.inventory.info.region_info import *
from spaceone.inventory.info.job_info import *
from spaceone.inventory.info.job_task_info import *
from spaceone.inventory.info.cloud_service_type_info import *
from spaceone.inventory.info.cloud_service_info import *
from spaceone.inventory.info.cloud_service_query_set_info import *
from spaceone.inventory.info.cloud_service_stats_info import *
from spaceone.inventory.info.collector_info import *
from spaceone.inventory.info.resource_group_info import *
from spaceone.inventory.info.change_history_info import *
from spaceone.inventory.info.note_info import *
from spaceone.inventory.info.collector_rule_info import *



import abc
from spaceone.inventory.error import *

# CREATED = 'CREATED'
INPROGRESS = 'IN_PROGRESS'
CANCELED = 'CANCELED'
SUCCESS = 'SUCCESS'
# ERROR = 'ERROR'
FAILURE = 'FAILURE'
# TIMEOUT = 'TIMEOUT'


class JobState(metaclass=abc.ABCMeta):
    def __init__(self):
        self.handle()

    @abc.abstractmethod
    def handle(self):
        pass


class InprogressState(JobState):
    def handle(self):
        pass

    def __str__(self):
        return INPROGRESS

# DEPRECATED
# class CreatedState(JobState):
#     def handle(self):
#         pass
#
#     def __str__(self):
#         return CREATED


class CanceledState(JobState):
    def handle(self):
        pass

    def __str__(self):
        return CANCELED


class SuccessState(JobState):
    def handle(self):
        pass

    def __str__(self):
        return SUCCESS

class FailureState(JobState):
    def handle(self):
        pass

    def __str__(self):
        return FAILURE

# DEPRECATED
# class ErrorState(JobState):
#     def handle(self):
#         pass
#
#     def __str__(self):
#         return ERROR


# DEPRECATED
# class TimeoutState(JobState):
#     def handle(self):
#         pass
#
#     def __str__(self):
#         return TIMEOUT


STATE_DIC = {
    # 'CREATED': CreatedState(),
    'IN_PROGRESS': InprogressState(),
    'CANCELED': CanceledState(),
    'SUCCESS': SuccessState(),
    'FAILURE': SuccessState(),
    # 'ERROR': ErrorState(),
    # 'TIMEOUT': TimeoutState()
}


class JobStateMachine():
    def __init__(self, job_vo):
        self.job_id = job_vo.job_id
        self._status = STATE_DIC[job_vo.status]

    def inprogress(self):
        if isinstance(self._status, (InprogressState, SuccessState)):
            self._status = InprogressState()
        elif isinstance(self._status, (FailureState,)):
            pass
        else:
            raise ERROR_JOB_STATE_CHANGE(action='INPROGRESS', job_id=self.job_id, status=str(self._status))
        return self.get_status()

    def canceled(self):
        if isinstance(self._status, InprogressState):
            self._status = CanceledState()
        else:
            raise ERROR_JOB_STATE_CHANGE(action='CANCELED', job_id=self.job_id, status=str(self._status))
        return self.get_status()

    def success(self):
        if isinstance(self._status, (InprogressState, SuccessState)):
            self._status = SuccessState()
        elif isinstance(self._status, (FailureState,)):
            pass
        else:
            raise ERROR_JOB_STATE_CHANGE(action='SUCCESS', job_id=self.job_id, status=str(self._status))
        return self.get_status()

    def failure(self):
        self._status = FailureState()
        return self.get_status()

    def get_status(self):
        return str(self._status)

    # DEPRECATED
    # def timeout(self):
    #     if isinstance(self._status, (CreatedState, InprogressState)):
    #         self._status = TimeoutState()
    #     else:
    #         raise ERROR_JOB_STATE_CHANGE(action='TIMEOUT', job_id=self.job_id, status=str(self._status))
    #     return self.get_status()

    # DEPRECATED
    def error(self):
        self._status = ErrorState()
        return self.get_status()


import logging

from spaceone.core.manager import BaseManager
from spaceone.inventory.lib.resource_manager import ResourceManager
from spaceone.inventory.model.resource_group_model import ResourceGroup

_LOGGER = logging.getLogger(__name__)


class ResourceGroupManager(BaseManager, ResourceManager):

    resource_keys = ['resource_group_id']
    query_method = 'list_resource_groups'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resource_group_model: ResourceGroup = self.locator.get_model('ResourceGroup')

    def create_resource_group(self, params):
        def _rollback(rg_vo):
            _LOGGER.info(f'[ROLLBACK] Delete resource group : {rg_vo.name} ({rg_vo.resource_group_id})')
            rg_vo.delete()

        rg_vo: ResourceGroup = self.resource_group_model.create(params)
        self.transaction.add_rollback(_rollback, rg_vo)

        return rg_vo

    def update_resource_group(self, params):
        return self.update_resource_group_by_vo(params, self.get_resource_group(params['resource_group_id'],
                                                                                params['domain_id']))

    def update_resource_group_by_vo(self, params, rg_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[ROLLBACK] Revert Data : {old_data["name"]} ({old_data["resource_group_id"]})')
            rg_vo.update(old_data)

        self.transaction.add_rollback(_rollback, rg_vo.to_dict())
        return rg_vo.update(params)

    def delete_resource_group(self, resource_group_id, domain_id):
        self.delete_resource_group_by_vo(self.get_resource_group(resource_group_id, domain_id))

    def get_resource_group(self, resource_group_id, domain_id, only=None):
        return self.resource_group_model.get(resource_group_id=resource_group_id, domain_id=domain_id, only=only)

    def list_resource_groups(self, query):
        return self.resource_group_model.query(**query)

    def stat_resource_groups(self, query):
        return self.resource_group_model.stat(**query)

    @staticmethod
    def delete_resource_group_by_vo(resource_group_vo):
        resource_group_vo.delete()

    def get_resource_group_filter(self, resource_group_id, resource_type, domain_id, keyword_filter):
        filters = []
        resource_group_vo = self.get_resource_group(resource_group_id, domain_id)
        for resource in resource_group_vo.resources:
            if resource.resource_type.startswith(resource_type):
                _filter = self._get_filter_from_resource_type(resource.resource_type)
                _filter += resource.filter

                _filter.append({
                    'k': 'domain_id',
                    'v': domain_id,
                    'o': 'eq'
                })

                if resource_group_vo.project_id:
                    _filter.append({
                        'k': 'project_id',
                        'v': resource_group_vo.project_id,
                        'o': 'eq'
                    })

                query = {
                    'filter': _filter
                }

                if resource.keyword and resource.keyword.strip() != '':
                    query['filter_or'] = self._get_resource_group_keyword_filter(resource.keyword, keyword_filter)

                filters.append(query)

        return filters

    @staticmethod
    def _get_resource_group_keyword_filter(keyword, keyword_filter):
        filter_or = []
        for key in keyword_filter:
            filter_or.append({
                'k': key,
                'v': list(filter(None, keyword.split(' '))),
                'o': 'contain_in'
            })

        return filter_or

    def _get_filter_from_resource_type(self, resource_type):
        resource_type_split = resource_type.split('?', 1)
        if len(resource_type_split) == 2:
            return self._get_filter_from_query_string(resource_type_split[1])
        else:
            return []

    @staticmethod
    def _get_filter_from_query_string(query_string):
        _filter = []
        for kv in query_string.split('&'):
            kv_split = kv.split('=', 1)
            if len(kv_split) == 2:
                if kv_split[0].strip() != '' and kv_split[1].strip() != '':
                    _filter.append({
                        'k': kv_split[0],
                        'v': kv_split[1],
                        'o': 'eq'
                    })

        return _filter



import logging
from spaceone.core.service import *
from spaceone.core import utils
from spaceone.inventory.manager.resource_group_manager import ResourceGroupManager
from spaceone.inventory.manager.identity_manager import IdentityManager

_LOGGER = logging.getLogger(__name__)
_KEYWORD_FILTER = ['resource_group_id', 'name']


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ResourceGroupService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.resource_group_mgr: ResourceGroupManager = self.locator.get_manager('ResourceGroupManager')

    @transaction(append_meta={
        'authorization.scope': 'PROJECT',
        'authorization.require_project_id': True
    })
    @check_required(['name', 'resources', 'domain_id'])
    def create(self, params):
        """
        Args:
            params (dict): {
                    'name': 'str',
                    'resources': 'list',
                    'project_id': 'str',
                    'options': 'dict',
                    'tags': 'dict',
                    'domain_id': 'str'
                }

        Returns:
            resource_group_vo (object)
        """

        if 'tags' in params:
            if isinstance(params['tags'], list):
                params['tags'] = utils.tags_to_dict(params['tags'])

        return self.resource_group_mgr.create_resource_group(params)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['resource_group_id', 'domain_id'])
    def update(self, params):
        """
        Args:
            params (dict): {
                    'resource_group_id': 'str',
                    'name': 'str',
                    'resources': 'list',
                    'project_id': 'str',
                    'release_project': 'bool',
                    'options': 'dict',
                    'tags': 'dict',
                    'domain_id': 'str'
                }

        Returns:
            resource_group_vo (object)

        """

        if 'tags' in params:
            if isinstance(params['tags'], list):
                params['tags'] = utils.tags_to_dict(params['tags'])

        identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
        project_id = params.get('project_id', self.transaction.get_meta('secret.project_id'))
        release_project = params.get('release_project', False)

        if release_project:
            params['project_id'] = None
        elif project_id:
            identity_mgr.get_project(project_id, params['domain_id'])
            params['project_id'] = project_id

        rg_vo = self.resource_group_mgr.get_resource_group(params['resource_group_id'], params['domain_id'])
        return self.resource_group_mgr.update_resource_group_by_vo(params, rg_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['resource_group_id', 'domain_id'])
    def delete(self, params):
        """
        Args:
            params (dict): {
                    'resource_group_id': 'str',
                    'domain_id': 'str'
                }

        Returns:
            None

        """

        rg_vo = self.resource_group_mgr.get_resource_group(params['resource_group_id'], params['domain_id'])
        self.resource_group_mgr.delete_resource_group_by_vo(rg_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['resource_group_id', 'domain_id'])
    def get(self, params):
        """
        Args:
            params (dict): {
                    'resource_group_id': 'str',
                    'domain_id': 'str',
                    'only': 'list'
                }

        Returns:
            resource_group_vo (object)

        """

        return self.resource_group_mgr.get_resource_group(params['resource_group_id'], params['domain_id'],
                                                          params.get('only'))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['domain_id'])
    @append_query_filter(['resource_group_id', 'name', 'project_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(_KEYWORD_FILTER)
    def list(self, params):
        """
        Args:
            params (dict): {
                    'resource_group_id': 'str',
                    'name': 'str',
                    'project_id': 'str',
                    'domain_id': 'str',
                    'query': 'dict (spaceone.api.core.v1.Query)',
                    'user_projects': 'list', // from meta
                }

        Returns:
            results (list)
            total_count (int)

        """
        return self.resource_group_mgr.list_resource_groups(params.get('query', {}))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id', 'user_projects'])
    @append_keyword_filter(_KEYWORD_FILTER)
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.resource_group_mgr.stat_resource_groups(query)



from mongoengine import *
from spaceone.core.model.mongo_model import MongoModel


class CollectionState(MongoModel):
    collector_id = StringField(max_length=40)
    job_task_id = StringField(max_length=40)
    secret_id = StringField(max_length=40)
    resource_id = StringField(max_length=40)
    resource_type = StringField(max_length=255)
    disconnected_count = IntField(default=0)
    domain_id = StringField(max_length=40)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        'updatable_fields': [
            'job_task_id',
            'disconnected_count',
            'updated_at'
        ],
        'indexes': [
            {
                "fields": ['domain_id', 'collector_id', 'resource_id', 'secret_id'],
                "name": "COMPOUND_INDEX_FOR_GET"
            },
            {
                "fields": ['domain_id', 'collector_id', '-disconnected_count'],
                "name": "COMPOUND_INDEX_FOR_DELETE_1"
            },
            {
                "fields": ['resource_id', 'domain_id'],
                "name": "COMPOUND_INDEX_FOR_DELETE_2"
            },
            {
                "fields": ['domain_id', 'collector_id', 'job_task_id', 'secret_id', 'updated_at'],
                "name": "COMPOUND_INDEX_FOR_DELETE_3"
            },
        ]
    }




import functools
import logging
from spaceone.api.inventory.v1 import resource_group_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.inventory.model.resource_group_model import ResourceGroup, Resource

__all__ = ['ResourceGroupInfo', 'ResourceGroupsInfo']

_LOGGER = logging.getLogger(__name__)


def ResourceInfo(resource: Resource):
    info = {
        'resource_type': resource.resource_type,
        'filter': change_list_value_type(resource.filter),
        'keyword': resource.keyword
    }
    return resource_group_pb2.Resource(**info)


def ResourceGroupInfo(rg_vo: ResourceGroup, minimal=False):
    info = {
        'resource_group_id': rg_vo.resource_group_id,
        'name': rg_vo.name,
        'project_id': rg_vo.project_id
    }

    if not minimal:
        info.update({
            'resources': list(map(ResourceInfo, rg_vo.resources)),
            'options': change_struct_type(rg_vo.options),
            'tags': change_struct_type(rg_vo.tags),
            'domain_id': rg_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(rg_vo.created_at),
        })

    return resource_group_pb2.ResourceGroupInfo(**info)


def ResourceGroupsInfo(rg_vos, total_count, **kwargs):
    return resource_group_pb2.ResourceGroupsInfo(results=list(map(functools.partial(ResourceGroupInfo, **kwargs),
                                                                  rg_vos)),
                                                 total_count=total_count)



"""
Deprecated
"""

import logging

from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.inventory.error import *
from spaceone.inventory.manager.collecting_manager import RESOURCE_MAP

_LOGGER = logging.getLogger(__name__)


class FilterManager(BaseManager):
    """
    Transform filter for collector
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.CONSTANT_FILTER_CACHE_TIMEOUT = 86400  # 24 hours

    #####################################################
    # TODO: result of _get_collect_filter, and secret_id
    """ I want to know the founded result from _get_collect_filter must be related with secret_id
    If resource is not collected by this secret_id, I don't want to make collect call
    """

    def get_collect_filter(self, filters, plugin_info, secret_id_list=[]):
        """ Create new filters for Collect plugin's parameter
            filter_format(filters) -> new_filter

        Args:
            filters(dict): filters from Client request
            plugin_info(dict)
            secret_id_list(list)

        Returns:
            new_filter: new filters for Plugin(Collector) query
            related_secret_id_list : list of secret matched on query

        Example:
            'filter_format': [
                {
                    'change_rules': [
                        {'change_key': 'instance_id', 'resource_key': 'data.compute.instance_id'},
                        {'change_key': 'region_name', 'resource_key': 'data.compute.region'}
                    ],
                    'key': 'project_id',
                    'name': 'Project ID',
                    'resource_type': 'SERVER',
                    'search_key': 'identity.Project.project_id',
                    'type': 'str'
                },
                {
                    'change_rules': [
                        {'change_key': 'instance_id', 'resource_key': 'data.compute.instance_id'},
                        {'change_key': 'region_name', 'resource_key': 'data.compute.region'}
                    ],
                    'key': 'collection_info.service_accounts',
                    'name': 'Service Account ID',
                    'resource_type': 'SERVER',
                    'search_key': 'identity.ServiceAccount.service_account_id',
                    'type': 'str'
                },
                {
                    'change_rules': [
                        {'change_key': 'instance_id', 'resource_key': 'data.compute.instance_id'},
                        {'change_key': 'region_name', 'resource_key': 'data.compute.region'}
                    ],
                    'key': 'server_id',
                    'name': 'Server ID',
                    'resource_type': 'SERVER',
                    'search_key': 'inventory.Server.server_id',
                    'type': 'list'
                },
                {
                    'key': 'instance_id',
                    'name': 'Instance ID',
                    'resource_type': 'CUSTOM',
                    'type': 'list'
                },
                {
                    'key': 'region_name',
                    'name': 'Region',
                    'resource_type': 'CUSTOM',
                    'type': 'list'}],

            filters:
                {
                    'region_id': 'region-xxxxx',
                    'zone_id': 'zone-yyyyy',
                    'instance_id': ['i-zzzz', ...]]            # CUSTOM resource type
                    'instance_type': 'm4.xlarge'
                }

            new_filter:
                {
                    'instance_id': ['i-123', 'i-2222', ...]
                    'instance_type': 'm4.xlarge'
                }
            related_secret_id_list: ['secret-12343', 'secret-254555' ...]
            :param filters:
            :param secret_id_list:
            :param plugin_info:

        """

        metadata = plugin_info.get('metadata', None)
        #############################
        # WARNING
        # options is old spec.
        #############################
        options = plugin_info.get('options', {})
        if metadata:
            options = metadata

        filter_format = options.get('filter_format', None)
        if filter_format is None:
            _LOGGER.warning(f'[_get_collector_filter] No filter_format at plugin_info')
            return {}, secret_id_list
        if filters == {}:
            _LOGGER.debug(f'[_get_collector_filter] No filters, do nothing')
            return {}, secret_id_list

        filter_format_by_key = {}
        # filter_format_by_key['zone_id'] = {'key':'project_id', 'name':'Project ID' ...
        for item in filter_format:
            filter_format_by_key[item['key']] = item

        for filter_key, filter_value in filters.items():
            if filter_key not in filter_format_by_key:
                _LOGGER.error(f'[_get_collect_filter] unsupported filter_key: {filter_key}')
                # Strict error raise, for reducing too heavy request
                raise ERROR_UNSUPPORTED_FILTER_KEY(key=filter_key, value=filter_value)

        query_filter, custom_keys = self._prepare_query_filter(filters, filter_format_by_key)
        _LOGGER.debug(f'[_get_collect_filter] query_filter: {query_filter}, custom_keys: {custom_keys}')

        query_per_resources = self._make_query_per_resources(query_filter, filter_format_by_key)
        _LOGGER.debug(f'[_get_collect_filter] query_per_resources: {query_per_resources}')

        new_filter, filtered_secret_id_list = self._search_resources(query_per_resources,
                                                                     filter_format_by_key,
                                                                     secret_id_list)
        _LOGGER.debug(f'[_get_collect_filter] new_filter: {new_filter}')
        related_secret_id_list = _intersection(secret_id_list, filtered_secret_id_list)

        if len(custom_keys) > 0:
            new_filter = self._append_custom_keys(new_filter, filters, custom_keys)
            _LOGGER.debug(f'[_get_collect_filter] new_filter_with_custom_keys: {new_filter}')

        return new_filter, related_secret_id_list

    def cache_filter(self, collector_id, secret_id, data):
        """
        FilerManager can save cache of filter for collect plugin
        Save the region_name cache
        Args:
            data (dict): {
                'region_name': list of region name,
                'cloud_service': list of cloud service (for cloud service plugin)
                }

        Key: collector-filter:<collector_id>:<secret_id>
        Value: region_name: [list of regions]
        """
        key = f'collector-filter:{collector_id}:{secret_id}'
        _LOGGER.debug(f'[cache_filter] {key} : {data}')
        cache.set(key, data, expire=self.CONSTANT_FILTER_CACHE_TIMEOUT)

    @staticmethod
    def _get_filer_cache(collector_id, secret_id):
        key = f'collector-filter:{collector_id}:{secret_id}'
        try:
            data = cache.get(key)
            _LOGGER.debug(f'[cache_filter] {key} : {data}')
            return data
        except Exception as e:
            # May be no_cache
            return None

    def _prepare_query_filter(self, filters, filter_format_by_key):
        query_filter = {}
        """
        'region_id': [{'k': 'region_id', 'v': 'region-xxx', 'o': 'eq'}]
        'server_id': [{'k': 'server_id', 'v': 'server-yyyy', 'o': 'eq'} ....]
        ...
        """
        custom_keys = {}
        # Foreach filter, we will find matched resource list
        for filter_key, filter_value in filters.items():
            # filter_key : region_id
            filter_element = filter_format_by_key[filter_key]
            _LOGGER.debug(f'[_prepare_query_filter] filter_element: {filter_element}')
            if filter_element['resource_type'] == 'CUSTOM':
                # DO NOT save CUSTOM key at query_filter
                custom_keys[filter_key] = filter_element
                continue

            # list of new_filter[key]
            v_list = query_filter.get(filter_key, [])
            if filter_element:
                # Ask to manager, is there any matched resource
                query = self._make_query_for_manager(filter_key, filter_value, filter_element)
                if isinstance(query, list) is False:
                    _LOGGER.error("LOGIC ERROR, _make_query_for_manager does not return list value: {query}")
                else:
                    v_list.extend(query)
            query_filter[filter_key] = v_list
        return query_filter, custom_keys

    @staticmethod
    def _make_query_per_resources(query_filter, filter_format_by_key):
        # Make query per Resource
        query_per_resources = {}
        """
        'SERVER': {
            'key': 'zone_id',
            'filter': [{'k': 'region_id', 'v': 'region-xxxx', 'o': 'eq'}],
            'filter_or': [{'k': 'server_id', 'v': 'server-yyyy', 'o': 'eq'}, ...]
            }
        """
        for query_key, query in query_filter.items():
            res_type = filter_format_by_key[query_key]['resource_type']
            query_string = query_per_resources.get(res_type, {'key': query_key, 'filter': [], 'filter_or': []})
            if len(query) == 1:
                query_string['filter'].extend(query)
            elif len(query) > 1:
                query_string['filter_or'].extend(query)
            else:
                _LOGGER.debug(f'[_get_collector_filter] wrong query: {query}')
            query_per_resources[res_type] = query_string
        return query_per_resources

    def _search_resources(self, query_per_resources, filter_format_by_key, secret_id_list):
        """
        # Search Resource by Resource's Manager

        Returns: tuple of transformed query, secret_id_list
                 transformed_query {
                    'instance_id': [list of value],
                }
                 related_secret_id_list : [list of secrets]
        """
        result = {}
        # secret_id_list = []
        for res_type, query in query_per_resources.items():
            """ Example
            query: {'key': 'zone_id',
                    'filter': [
                            {'k': 'zone_id', 'v': 'zone-d710c1cb0ea7', 'o': 'eq'},
                            {'k': 'region_id', 'v': 'region-85445849c20c', 'o': 'eq'},
                            {'k': 'pool_id', 'v': 'pool-a1f35b107bb4', 'o': 'eq'}],
                    'filter_or': []}
            """
            _LOGGER.debug(f'[_search_resources] query: {query}')
            try:
                mgr = self.locator.get_manager(RESOURCE_MAP[res_type][1])
            except Exception as e:
                _LOGGER.error('########## NOTICE to Developer (bug) ###################################')
                _LOGGER.error(f'[_search_resources] Not found manager based on resource_type: {res_type}')
                _LOGGER.error(e)
                continue

            """
            {'change_rules': [{'change_key': 'instance_id',
                              'resource_key': 'data.compute.instance_id'},
                             {'change_key': 'region_name',
                              'resource_key': 'data.compute.region'}],
            """
            filter_element = filter_format_by_key[query['key']]
            change_rules = filter_element['change_rules']
            del query['key']

            # Ask to manager
            try:
                _LOGGER.debug(f'[_search_resources] query: {query}, key={change_rules}')
                value_list, filtered_secret_id_list = mgr.query_resources(query, change_rules)
                _LOGGER.debug(f'[_search_resources] filtered: {value_list}')
                result.update(value_list)
                secret_id_list = _intersection(secret_id_list, filtered_secret_id_list)
            except Exception as e:
                _LOGGER.error('########## NOTICE to Developer (bug) ####################################')
                _LOGGER.error(f'{res_type} Manager has bug for query_resources functions')
                _LOGGER.error(e)

        return result, secret_id_list

    @staticmethod
    def _append_custom_keys(new_filter, filters, custom_keys):
        """
        Args: {'key':'instance_id', 'name':'Instance ID', 'type':'list', 'resource_type': 'CUSTOM'}

        Return: updated new_filter
        """
        updated_filter = new_filter.copy()
        for custom_key, formats in custom_keys.items():
            _LOGGER.debug(f'[_append_custom_keys] append custom_key: {custom_key}, {formats}')
            values = filters.get(custom_key, None)
            if values is None:
                continue
            value_type = formats['type']
            _LOGGER.debug(f'[_append_custom_keys] find values: {values}, type: {value_type}')
            if value_type == 'list':
                current_value = new_filter.get(custom_key, [])
                current_value.extend(values)
                updated_filter.update({custom_key: current_value})
            elif value_type == 'str':
                current_value = new_filter.get(custom_key, None)
                if current_value:
                    _LOGGER.warning(f'[_append_custom_keys] change filter_value: {current_value} -> {values}')
                updated_filter.update({custom_key: values})
            else:
                _LOGGER.error(f'[_append_custom_keys] un-supported type: {formats}, type: {value_type}')
        _LOGGER.debug(f'[_append_custom_keys] updated_filter: {updated_filter}')
        return updated_filter

    @staticmethod
    def _make_query_for_manager(key, value, filter_element):
        """
        Args:
            key(str): key for query
            value: query value of element (str, int, bool, float, list)
            filter_element(dict): one element for filter_format

        Returns:
            query_statement (list, since there are list type)

        Example)
            value: region-xxxxx
            filter_element: {'key':'region_id', 'name':'Region', 'type':'str', 'resource_type': 'SERVER', 'change_key': ['data.compute.instance_id', 'instance_id']}
        """
        query_filter = []

        f_type = filter_element['type']
        if f_type == 'list':
            query_filter.append({
                'k': key,
                'v': value,
                'o': 'in'
            })
        elif f_type == 'str':
            query_filter.append({
                'k': key,
                'v': value,
                'o': 'eq'
            })
        else:
            _LOGGER.error(f'Unsupported filter_element, {filter_element}, supported type: list | str')
        return query_filter


def _intersection(list_a, list_b):
    """ Return intersection between list_a and list_b
    """
    if len(list_b) == 0:
        # May be user send blank list
        return list_a

    a = set(list_a)
    b = set(list_b)
    c = a.intersection(b)
    _LOGGER.debug(f'[_intersection] a: {list_a}, b: {list_b} -> {c}')
    return list(c)



import logging
from datetime import datetime

from spaceone.core import config
from spaceone.core import utils
from spaceone.core.token import get_token
from spaceone.core.locator import Locator
from spaceone.core.scheduler import HourlyScheduler

_LOGGER = logging.getLogger(__name__)


class CloudServiceStatsScheduler(HourlyScheduler):

    def __init__(self, queue, interval, minute=':00'):
        super().__init__(queue, interval, minute)
        self.locator = Locator()
        self._init_config()
        self._create_metadata()

    def _init_config(self):
        self._token = get_token('TOKEN')
        self._stats_schedule_hour = config.get_global('STATS_SCHEDULE_HOUR', 16)

    def _create_metadata(self):
        self._metadata = {
            'service': 'inventory',
            'resource': 'CloudServiceQuerySet',
            'verb': 'run_all_query_sets',
            'token': self._token,
        }

    def create_task(self):
        if datetime.utcnow().hour == self._stats_schedule_hour:
            stp = {
                'name': 'cloud_service_stats_schedule',
                'version': 'v1',
                'executionEngine': 'BaseWorker',
                'stages': [{
                    'locator': 'SERVICE',
                    'name': 'CloudServiceQuerySetService',
                    'metadata': self._metadata,
                    'method': 'run_all_query_sets',
                    'params': {
                        'params': {}
                    }
                }]
            }

            print(f'{utils.datetime_to_iso8601(datetime.now())} '
                  f'[INFO] [create_task] run_all_query_sets => START')
            return [stp]
        else:
            print(f'{utils.datetime_to_iso8601(datetime.now())} '
                  f'[INFO] [create_task] run_all_query_sets => SKIP')
            print(f'{utils.datetime_to_iso8601(datetime.now())} [INFO] '
                  f'[create_task] data_source_sync_time: {self._stats_schedule_hour} hour (UTC)')
            return []



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class RecordDiff(EmbeddedDocument):
    key = StringField(required=True)
    before = DynamicField(default=None, null=True)
    after = DynamicField(default=None, null=True)
    type = StringField(max_length=20, choices=('ADDED', 'CHANGED', 'DELETED'), required=True)

    def to_dict(self):
        return dict(self.to_mongo())


class Record(MongoModel):
    record_id = StringField(max_length=40, generate_id='record', unique=True)
    cloud_service_id = StringField(max_length=40, required=True)
    action = StringField(max_length=20, choices=('CREATE', 'UPDATE', 'DELETE'), required=True)
    diff = ListField(EmbeddedDocumentField(RecordDiff), default=[])
    diff_count = IntField(default=0)
    user_id = StringField(max_length=255, default=None, null=True)
    collector_id = StringField(max_length=40, default=None, null=True)
    job_id = StringField(max_length=40, default=None, null=True)
    updated_by = StringField(max_length=40, choices=('COLLECTOR', 'USER'), default=None, null=True)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now=True)

    meta = {
        'minimal_fields': [
            'record_id',
            'cloud_service_id',
            'action',
            'diff_count',
            'user_id',
            'collector_id',
            'job_id',
            'updated_by',
            'created_at'
        ],
        'ordering': [
            '-created_at'
        ],
        'indexes': [
            {
                "fields": ['domain_id', 'cloud_service_id', '-created_at', 'diff.key'],
                "name": "COMPOUND_INDEX_FOR_SEARCH"
            },
            {
                "fields": ['domain_id', 'record_id'],
                "name": "COMPOUND_INDEX_FOR_GET"
            }
        ]
    }



from spaceone.api.inventory.v1 import cloud_service_pb2, cloud_service_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class CloudService(BaseAPI, cloud_service_pb2_grpc.CloudServiceServicer):

    pb2 = cloud_service_pb2
    pb2_grpc = cloud_service_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceService', metadata) as cloud_svc_service:
            return self.locator.get_info('CloudServiceInfo', cloud_svc_service.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceService', metadata) as cloud_svc_service:
            return self.locator.get_info('CloudServiceInfo', cloud_svc_service.update(params))

    def pin_data(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceService', metadata) as cloud_svc_service:
            return self.locator.get_info('CloudServiceInfo', cloud_svc_service.pin_data(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceService', metadata) as cloud_svc_service:
            cloud_svc_service.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceService', metadata) as cloud_svc_service:
            return self.locator.get_info('CloudServiceInfo', cloud_svc_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceService', metadata) as cloud_svc_service:
            cloud_svc_vos, total_count = cloud_svc_service.list(params)
            return self.locator.get_info('CloudServicesInfo',
                                         cloud_svc_vos,
                                         total_count,
                                         minimal=self.get_minimal(params),
                                         include_metadata=False)

    def analyze(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceService', metadata) as cloud_svc_service:
            return self.locator.get_info('AnalyzeInfo', cloud_svc_service.analyze(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CloudServiceService', metadata) as cloud_svc_service:
            return self.locator.get_info('StatisticsInfo', cloud_svc_service.stat(params))



import functools
from spaceone.api.inventory.v1 import cloud_service_stats_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.inventory.model.cloud_service_stats_model import CloudServiceStats

__all__ = ['CloudServiceStatInfo', 'CloudServiceStatsInfo']


def CloudServiceStatInfo(cloud_svc_stat_vo: CloudServiceStats, minimal=False):
    info = {
        'query_set_id': cloud_svc_stat_vo.query_set_id,
        'values': change_struct_type(cloud_svc_stat_vo.values),
        'unit': change_struct_type(cloud_svc_stat_vo.unit),
        'provider': cloud_svc_stat_vo.provider,
        'cloud_service_group': cloud_svc_stat_vo.cloud_service_group,
        'cloud_service_type': cloud_svc_stat_vo.cloud_service_type,
        'project_id': cloud_svc_stat_vo.project_id,
        'created_date': cloud_svc_stat_vo.created_date,
    }

    if not minimal:
        info.update({
            'region_code': cloud_svc_stat_vo.region_code,
            'account': cloud_svc_stat_vo.account,
            'additional_info': change_struct_type(cloud_svc_stat_vo.additional_info),
            'domain_id': cloud_svc_stat_vo.domain_id
        })

    return cloud_service_stats_pb2.CloudServiceStatInfo(**info)


def CloudServiceStatsInfo(cloud_svc_stats_vos, total_count, **kwargs):
    return cloud_service_stats_pb2.CloudServiceStatsInfo(
        results=list(map(functools.partial(CloudServiceStatInfo, **kwargs), cloud_svc_stats_vos)),
        total_count=total_count)



import functools
from typing import List
from spaceone.api.inventory.v1 import cloud_service_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.inventory.model.cloud_service_model import CloudService

__all__ = ['CloudServiceInfo', 'CloudServicesInfo']


def CollectionInfo(vos: list):
    if vos:
        collections = []
        for vo in vos:
            info = {
                'provider': vo.provider,
                'service_account_id': vo.service_account_id,
                'secret_id': vo.secret_id,
                'collector_id': vo.collector_id,
                'last_collected_at': utils.datetime_to_iso8601(vo.last_collected_at)
            }
            collection = cloud_service_pb2.CollectionInfo(**info)
            collections.append(collection)
        return collections
    else:
        return None


def CloudServiceInfo(cloud_svc_vo: CloudService, minimal=False, include_metadata=True):
    info = {
        'cloud_service_id': cloud_svc_vo.cloud_service_id,
        'name': cloud_svc_vo.name,
        'state': cloud_svc_vo.state,
        'cloud_service_group': cloud_svc_vo.cloud_service_group,
        'cloud_service_type': cloud_svc_vo.cloud_service_type,
        'provider': cloud_svc_vo.provider,
        'region_code': cloud_svc_vo.region_code,
        'reference': cloud_service_pb2.CloudServiceReference(
            **cloud_svc_vo.reference.to_dict()) if cloud_svc_vo.reference else None,
        'project_id': cloud_svc_vo.project_id,
    }

    if not minimal:
        info.update({
            'account': cloud_svc_vo.account,
            'instance_type': cloud_svc_vo.instance_type,
            'instance_size': cloud_svc_vo.instance_size,
            'ip_addresses': cloud_svc_vo.ip_addresses,
            'data': change_struct_type(cloud_svc_vo.data),
            'tags': change_struct_type(_change_tags_without_hash(cloud_svc_vo.tags)),
            'tag_keys': change_struct_type(cloud_svc_vo.tag_keys),
            'collection_info': CollectionInfo(cloud_svc_vo.collection_info),
            'domain_id': cloud_svc_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(cloud_svc_vo.created_at),
            'updated_at': utils.datetime_to_iso8601(cloud_svc_vo.updated_at),
            'deleted_at': utils.datetime_to_iso8601(cloud_svc_vo.deleted_at),
        })

        if include_metadata:
            info.update({
                'metadata': change_struct_type(cloud_svc_vo.metadata),
            })

    return cloud_service_pb2.CloudServiceInfo(**info)


def CloudServicesInfo(cloud_svc_vos, total_count, **kwargs):
    return cloud_service_pb2.CloudServicesInfo(results=list(
        map(functools.partial(CloudServiceInfo, **kwargs), cloud_svc_vos)), total_count=total_count)


def _change_tags_without_hash(tags) -> dict:
    changed_tags = {}
    for provider, hashed_tags in tags.items():
        for hash_key, tag in hashed_tags.items():
            if not changed_tags.get(provider):
                changed_tags[provider] = {}
            changed_tags[provider][tag['key']] = tag['value']

    return changed_tags



"""
DEPRECATED
"""
import logging
import time
import schedule
import json
from jsonschema import validate
from spaceone.core import queue
from spaceone.core.locator import Locator
from spaceone.core.scheduler import IntervalScheduler
from spaceone.core.scheduler.task_schema import SPACEONE_TASK_SCHEMA
from spaceone.inventory.lib.scheduler import init_count, update_token, get_domain_id_from_token, get_interval_value
from spaceone.inventory.service.collector_service import CollectorService

_LOGGER = logging.getLogger(__name__)


"""
IntervalExecutor is a sync engine, for IntervalScheduler
Sync DB in every N seconds.

Load all interval schedule, and create IntervalScheduler
"""
class InventoryIntervalScheduler(IntervalScheduler):

    def __init__(self, queue, interval):
        super().__init__(queue, interval)
        self.count = init_count()
        self.locator = Locator()
        self.TOKEN = update_token()
        self.domain_id = get_domain_id_from_token(self.TOKEN)
        self.schedule_info = {}
        self.idx = 0

    def run(self):
        # Every specific interval, check schedule
        schedule.every(self.config).seconds.do(self._check_interval)
        while True:
            schedule.run_pending()
            time.sleep(1)

    def push_task(self, schedule_id, interval_info):
        task = self._create_job_request(schedule_id, interval_info)
        _LOGGER.debug(f'[push_task] {task["name"]}')
        try:
            validate(task, schema=SPACEONE_TASK_SCHEMA)
            queue.put(self.queue, json.dumps(task))
        except Exception as e:
            print(e)
            _LOGGER.debug(f'[push_task] Task schema: {task}, {e}')

    def _check_interval(self):
        # patch schedule and sync
        # create new interval Schedule per schedule_id
        # interval_schedules = {schedule-zzz: {interval: 20, domain_id: domain-1234, collector_id: collector-5678},
        #                       schedule_yyy: {interval: 30, domain_id: domain-2222, collector_id: collector-3333}}

        interval_schedules = self._get_interval_schedules()
        schedule_ids = []

        for schedule_id, interval_info in interval_schedules.items():
            schedule_ids.append(schedule_id)
            try:
                interval_value = interval_info['interval']
                # Create New scheduler
                if schedule_id not in self.schedule_info:
                    _LOGGER.debug(f'[_check_interval] create {schedule_id}, at every {interval_value}')
                    self.schedule_info[schedule_id] = interval_info
                    job = schedule.every(interval_value).seconds.do(self.push_task, schedule_id, interval_info)
                    # Add tag for each job
                    job.tag(schedule_id)
                # Sync previous scheduler
                else:
                    previous_interval_info = self.schedule_info[schedule_id]
                    previous_interval = previous_interval_info['interval']
                    if interval_value != previous_interval:
                        _LOGGER.debug(f'[_check_interval] modify {schedule_id} interval {previous_interval} to {interval_value}')
                        # delete job and re-create
                        schedule.default_scheduler.clear(tag=schedule_id)
                        job = schedule.every(interval_value).seconds.do(self.push_task, schedule_id, interval_info)
                        job.tag(schedule_id)
                        # Update self.schedule_info
                        self.schedule_info[schedule_id] = interval_info
                    else:
                        _LOGGER.debug(f'[_check_interval] continue {schedule_id}, at every {previous_interval}')
            except Exception as e:
                _LOGGER.error(f'[_check_interval] contact to developer {e}')

        # Delete garbage
        _LOGGER.debug(f'[_check_interval] gargabe collector: {len(schedule.default_scheduler.jobs) - 1}')
        for job in schedule.default_scheduler.jobs:
            if job.tags == set():
                continue
            exist = False
            for schedule_id in schedule_ids:
                if schedule_id in job.tags:
                    print(f'exist: {schedule_id}')
                    exist = True
                    break

            if exist is False:
                _LOGGER.debug(f'[_check_interval] remove job: {job}')
                schedule.default_scheduler.cancel_job(job)

    def _get_interval_schedules(self):
        found_schedules = {}

        for schedule_vo in self._list_schedules():
            try:
                interval_value = get_interval_value(schedule_vo)
                found_schedules.update({
                    schedule_vo.schedule_id: {
                        'interval': interval_value,
                        'domain_id': schedule_vo.domain_id,
                        'collector_id': schedule_vo.collector.collector_id
                    }
                })
            except Exception as e:
                _LOGGER.error(f'[_get_interval_schedules] {e}')

        _LOGGER.debug(f'[_get_interval_schedules] found: {found_schedules}')
        return found_schedules

    def _list_schedules(self):
        try:
            collector_svc: CollectorService = self.locator.get_service(CollectorService)
            schedule_vos, total = collector_svc.scheduled_collectors({'schedule': {'interval': None}})
            return schedule_vos
        except Exception as e:
            _LOGGER.error(e)
            return []

    def _create_job_request(self, schedule_id, interval_info):
        _LOGGER.debug(f'[_create_job_request] {interval_info}')
        domain_id = interval_info['domain_id']

        metadata = {
            'token': self.TOKEN,
            'service': 'inventory',
            'resource': 'Collector',
            'verb': 'collect',
            'domain_id': self.domain_id
        }

        schedule_job = {
            'locator': 'SERVICE',
            'name': 'CollectorService',
            'metadata': metadata,
            'method': 'collect',
            'params': {
                'params': {
                    'collector_id': interval_info['collector_id'],
                    'collect_mode': 'ALL',
                    'filter': {},
                    'domain_id': domain_id
                }
            }
        }

        return {
            'name': 'inventory_collect_by_interval_schedule',
            'version': 'v1',
            'executionEngine': 'BaseWorker',
            'stages': [schedule_job]
        }

    # @staticmethod
    # def _init_count():
    #     # get current time
    #     cur = datetime.datetime.utcnow()
    #     count = {
    #         'previous': cur,            # Last check_count time
    #         'index': 0,                # index
    #         'hour': cur.hour,           # previous hour
    #         'started_at': 0,            # start time of push_token
    #         'ended_at': 0               # end time of execution in this tick
    #         }
    #     _LOGGER.debug(f'[_init_count] {count}')
    #     return count
    #
    # @staticmethod
    # def _update_token():
    #     token = config.get_global('TOKEN')
    #     if token == "":
    #         token = _validate_token(config.get_global('TOKEN_INFO'))
    #     return token
    #
    # @staticmethod
    # def _get_domain_id_from_token(token):
    #     decoded_token = JWTUtil.unverified_decode(token)
    #     return decoded_token['did']
    #
    # @staticmethod
    # def _get_interval_value(schedule_vo):
    #     # get interval vaule from schedule vo
    #     return schedule_vo.schedule.interval


# class Consul:
#     def __init__(self, config):
#         """
#         Args:
#           - config: connection parameter
#
#         Example:
#             config = {
#                     'host': 'consul.example.com',
#                     'port': 8500
#                 }
#         """
#         self.config = self._validate_config(config)
#
#     def _validate_config(self, config):
#         """
#         Parameter for Consul
#         - host, port=8500, token=None, scheme=http, consistency=default, dc=None, verify=True, cert=None
#         """
#         options = ['host', 'port', 'token', 'scheme', 'consistency', 'dc', 'verify', 'cert']
#         result = {}
#         for item in options:
#             value = config.get(item, None)
#             if value:
#                 result[item] = value
#         return result
#
#     def patch_token(self, key):
#         """
#         Args:
#             key: Query key (ex. /debug/supervisor/TOKEN)
#
#         """
#         try:
#             conn = consul.Consul(**self.config)
#             index, data = conn.kv.get(key)
#             return data['Value'].decode('ascii')
#
#         except Exception as e:
#             _LOGGER.debug(f'[patch_token] failed: {e}')
#             return False




import logging
import functools
from spaceone.api.inventory.v1 import collector_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils

__all__ = ['CollectorInfo', 'CollectorsInfo', 'VerifyInfo']

_LOGGER = logging.getLogger(__name__)


def ScheduledInfo(schedule_vo):
    if schedule_vo:
        info = {
            'state': schedule_vo.state,
            'hours': schedule_vo.hours
        }
        return collector_pb2.Scheduled(**info)
    else:
        return None


def SecretFilterInfo(secret_filter_vo):
    if secret_filter_vo:
        info = {
            'state': secret_filter_vo.state,
            'secrets': secret_filter_vo.secrets,
            'service_accounts': secret_filter_vo.service_accounts,
            'schemas': secret_filter_vo.schemas,
            'exclude_secrets': secret_filter_vo.exclude_secrets,
            'exclude_service_accounts': secret_filter_vo.exclude_service_accounts,
            'exclude_schemas': secret_filter_vo.exclude_schemas
        }
        return collector_pb2.SecretFilter(**info)
    else:
        return None


def PluginInfo(vo, minimal=False):
    if vo is None:
        return None

    info = {
        'plugin_id': vo.plugin_id,
        'version': vo.version,
    }

    if not minimal:
        info.update({
            'options': change_struct_type(vo.options),
            'metadata': change_struct_type(vo.metadata),
            'upgrade_mode': vo.upgrade_mode
        })
    return collector_pb2.PluginInfo(**info)


def CollectorInfo(vo, minimal=False):
    info = {
        'collector_id': vo.collector_id,
        'name': vo.name,
        'provider': vo.provider,
        'capability': change_struct_type(vo.capability),
        'plugin_info': PluginInfo(vo.plugin_info, minimal=minimal),
    }

    if not minimal:
        info.update({
            'schedule': ScheduledInfo(vo.schedule),
            'secret_filter': SecretFilterInfo(vo.secret_filter),
            'created_at': utils.datetime_to_iso8601(vo.created_at),
            'last_collected_at': utils.datetime_to_iso8601(vo.last_collected_at),
            'tags': change_struct_type(vo.tags),
            'domain_id': vo.domain_id
        })

    return collector_pb2.CollectorInfo(**info)


def VerifyInfo(info):
    return collector_pb2.VerifyInfo(**info)


def CollectorsInfo(vos, total_count, **kwargs):
    return collector_pb2.CollectorsInfo(results=list(map(functools.partial(CollectorInfo, **kwargs), vos)),
                                        total_count=total_count)




from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel
from spaceone.inventory.model.cloud_service_type_model import CloudServiceType


class CloudServiceQuerySet(MongoModel):
    query_set_id = StringField(max_length=40, generate_id='query-set', unique=True)
    name = StringField(max_length=255, unique_with=['provider', 'cloud_service_group', 'cloud_service_type', 'domain_id'])
    state = StringField(max_length=20, default='ENABLED', choices=('ENABLED', 'DISABLED'))
    query_options = DictField()
    query_hash = StringField(max_length=255)
    query_type = StringField(max_length=20, required=True, choices=('MANAGED', 'CUSTOM'))
    unit = DictField()
    keys = ListField(StringField(max_length=255))
    additional_info_keys = ListField(StringField(max_length=255))
    provider = StringField(max_length=255, default=None, null=True)
    cloud_service_group = StringField(max_length=255, default=None, null=True)
    cloud_service_type = StringField(max_length=255, default=None, null=True)
    ref_cloud_service_type = StringField(max_length=255, default=None, null=True)
    tags = DictField()
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        'updatable_fields': [
            'name',
            'state',
            'query_options',
            'query_hash',
            'unit',
            'keys',
            'additional_info_keys',
            'tags',
            'updated_at'
        ],
        'minimal_fields': [
            'query_set_id',
            'name',
            'state',
            'query_type',
            'provider',
            'cloud_service_group',
            'cloud_service_type'
        ],
        'ordering': [
            'provider',
            'cloud_service_group',
            'cloud_service_type'
        ],
        'reference_query_keys': {
            'ref_cloud_service_type': {
                'model': CloudServiceType,
                'foreign_key': 'ref_cloud_service_type'
            }
        },
        'indexes': [
            'name',
            'state',
            'query_hash',
            'query_type',
            'provider',
            'cloud_service_group',
            'cloud_service_type',
            'domain_id'
        ]
    }



import functools
from spaceone.api.inventory.v1 import collector_pb2, job_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.inventory.model.job_model import Job
from spaceone.inventory.info.collector_info import CollectorInfo

__all__ = ['JobInfo', 'JobsInfo']


def ErrorInfo(error):
    info = {
        'error_code': error.error_code,
        'message': error.message,
        'additional': change_struct_type(error.additional)
    }
    return collector_pb2.ErrorInfo(**info)


def JobInfo(job_vo: Job, minimal=False):
    info = {
        'job_id': job_vo.job_id,
        'status': job_vo.status,
        'collector_id': job_vo.collector_id,
        'plugin_id': job_vo.plugin_id,
        'created_at': utils.datetime_to_iso8601(job_vo.created_at),
        'finished_at': utils.datetime_to_iso8601(job_vo.finished_at),
    }

    if not minimal:
        info.update({
            'secret_id': job_vo.secret_id,
            'total_tasks': job_vo.total_tasks,
            'remained_tasks': job_vo.remained_tasks,
            'success_tasks': job_vo.success_tasks,
            'failure_tasks': job_vo.failure_tasks,
            'domain_id': job_vo.domain_id,
            'updated_at': utils.datetime_to_iso8601(job_vo.updated_at),
        })

        # Temporary code for DB migration
        # if not job_vo.collector_id and job_vo.collector:
        #    job_vo.update({'collector_id': job_vo.collector.collector_id})

    return collector_pb2.JobInfo(**info)


def JobsInfo(vos, total_count, **kwargs):
    return job_pb2.JobsInfo(results=list(map(functools.partial(JobInfo, **kwargs), vos)), total_count=total_count)




import logging
import re

from datetime import datetime, timedelta

from spaceone.core import config
from spaceone.core.manager import BaseManager
from spaceone.inventory.manager.collection_state_manager import CollectionStateManager
from spaceone.inventory.manager.cloud_service_manager import CloudServiceManager
from spaceone.inventory.conf.collector_conf import *

_LOGGER = logging.getLogger(__name__)


class CleanupManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _delete_resources_by_collector(self, state_mgr, collector_id, domain_id):
        disconnected_count = config.get_global('DEFAULT_DISCONNECTED_STATE_DELETE_POLICY', 3)

        query = {
            'filter': [
                {'k': 'collector_id', 'v': collector_id, 'o': 'eq'},
                {'k': 'disconnected_count', 'v': disconnected_count, 'o': 'gte'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
            ]
        }

        state_vos, total_count = state_mgr.list_collection_states(query)
        cloud_service_ids = [state_vo.resource_id for state_vo in state_vos
                             if state_vo.resource_type == 'inventory.CloudService']

        total_deleted_count = 0

        if cloud_service_ids:
            cloud_svc_mgr: CloudServiceManager = self.locator.get_manager(CloudServiceManager)

            query = {
                'filter': [
                    {'k': 'cloud_service_id', 'v': cloud_service_ids, 'o': 'in'},
                    {'k': 'domain_id', 'v': domain_id, 'o': 'eq'}
                ]
            }

            try:
                deleted_count = cloud_svc_mgr.delete_resources(query)
                _LOGGER.debug(f'[_delete_resources_by_collector] delete cloud service {deleted_count} in {domain_id}')
                total_deleted_count = deleted_count
            except Exception as e:
                _LOGGER.error(f'[_delete_resources_by_collector] delete cloud service error: {e}', exc_info=True)

        return total_deleted_count

    def update_disconnected_and_deleted_count(self, collector_id, secret_id, job_task_id, domain_id):
        state_mgr: CollectionStateManager = self.locator.get_manager(CollectionStateManager)
        disconnected_count = self._increment_disconnected_count_by_collector(state_mgr, collector_id, secret_id,
                                                                             job_task_id, domain_id)
        deleted_count = self._delete_resources_by_collector(state_mgr, collector_id, domain_id)

        return (disconnected_count - deleted_count), deleted_count

    def delete_resources_by_policy(self, resource_type, hour, domain_id):
        updated_at = datetime.utcnow() - timedelta(hours=hour)
        resource_type_name, _filter = self._parse_resource_type(resource_type)

        _filter.extend([
            {'k': 'updated_at', 'v': updated_at, 'o': 'lt'},
            {'k': 'domain_id',  'v': domain_id, 'o': 'eq'}
        ])

        if resource_type_name in ['inventory.CloudServiceType', 'inventory.Region']:
            _filter.append({'k': 'updated_by', 'v': 'manual', 'o': 'not'})

        query = {'filter': _filter}

        _LOGGER.debug(f'[delete_resources_by_policy] query: {query}')
        if resource_type_name not in RESOURCE_MAP:
            _LOGGER.error(f'[delete_resources_by_policy] not found {resource_type_name}')
            return 0

        resource_manager = self.locator.get_manager(RESOURCE_MAP[resource_type_name][1])

        try:
            deleted_count = resource_manager.delete_resources(query)
            _LOGGER.debug(f'[delete_resources_by_policy] {deleted_count} in {domain_id}')
            return deleted_count

        except Exception as e:
            _LOGGER.error(f'[delete_resources] {e}', exc_info=True)
            return 0

    @staticmethod
    def _increment_disconnected_count_by_collector(state_mgr, collector_id, secret_id, job_task_id, domain_id):
        updated_at = datetime.utcnow() - timedelta(hours=1)

        query = {
            'filter': [
                {'k': 'collector_id', 'v': collector_id, 'o': 'eq'},
                {'k': 'secret_id', 'v': secret_id, 'o': 'eq'},
                {'k': 'job_task_id', 'v': job_task_id, 'o': 'not'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
                {'k': 'updated_at', 'v': updated_at, 'o': 'lt'},
            ]
        }

        state_vos, total_count = state_mgr.list_collection_states(query)
        state_vos.increment('disconnected_count')

        return total_count

    @staticmethod
    def _parse_resource_type(resource_type):
        """
        Example of resource type
         - inventory.CloudService
         - inventory.CloudService?provider=aws&cloud_service_group=DynamoDB
        """
        items = resource_type.split('?')
        resource_type_name = items[0]
        _filter = []
        if len(items) == 2:
            kv = items[1]
            kv_items = kv.split('&')
            for kv_item in kv_items:
                item = re.compile(r'(?P<k>[\w.]+)(?P<o>=|!=|<=|<|>=|>)(?P<v>[\w.]*)')
                m = item.match(kv_item)
                kvo = m.groupdict()
                if kvo['v'] == '':
                    value = None
                else:
                    value = kvo['v']
                _filter.append({'k': kvo['k'], 'o': OP_MAP[kvo["o"]], 'v': value})
                # item = kv_item.split('=')
                # if len(item) == 2:
                #    filter_list.append((item[0], item[1]))
        elif len(items) > 2:
            _LOGGER.error(f'[_parse_resource_type] wrong format: {resource_type}')

        _LOGGER.debug(f'[_parse_resource_type] {resource_type} -> {resource_type_name}, {_filter}')
        return resource_type_name, _filter



from spaceone.core.service import *
from spaceone.inventory.model.job_task_model import JobTask
from spaceone.inventory.manager.job_task_manager import JobTaskManager


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class JobTaskService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.job_task_mgr: JobTaskManager = self.locator.get_manager('JobTaskManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['job_task_id', 'domain_id'])
    def delete(self, params):
        """
        Args:
            params (dict): {
                    'job_task_id': 'str',
                    'domain_id': 'str'
                }

        Returns:
            None

        """
        job_task_id = params['job_task_id']
        domain_id = params['domain_id']

        job_task_vo: JobTask = self.job_task_mgr.get(job_task_id, domain_id)
        self.job_task_mgr.delete_job_task_by_vo(job_task_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['job_task_id', 'domain_id'])
    def get(self, params):
        """
        Args:
            params (dict): {
                    'job_task_id': 'str',
                    'domain_id': 'str',
                    'only': 'list'
                }

        Returns:
            job_task_vo (object)

        """

        return self.job_task_mgr.get(params['job_task_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['domain_id'])
    @append_query_filter(['job_task_id', 'status', 'job_id', 'secret_id', 'provider',
                          'service_account_id', 'project_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(['job_task_id'])
    def list(self, params):
        """
        Args:
            params (dict): {
                    'job_task_id': 'str',
                    'status': 'str',
                    'job_id': 'str',
                    'secret_id': 'str',
                    'provider': 'str',
                    'service_account_id': 'str',
                    'project_id': 'str',
                    'domain_id  ': 'str',
                    'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                    'user_projects': 'list', // from meta
                }

        Returns:
            results (list)
            total_count (int)

        """

        return self.job_task_mgr.list(params.get('query', {}))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id', 'user_projects'])
    @append_keyword_filter(['job_task_id'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.job_task_mgr.stat(query)



import functools
from spaceone.api.inventory.v1 import note_pb2
from spaceone.core import utils
from spaceone.inventory.model.note_model import Note

__all__ = ['NoteInfo', 'NotesInfo']


def NoteInfo(note_vo: Note, minimal=False):
    info = {
        'note_id': note_vo.note_id,
        'note': note_vo.note,
        'record_id': note_vo.record_id,
        'cloud_service_id': note_vo.cloud_service_id,
        'created_by': note_vo.created_by,
    }

    if not minimal:
        info.update({
            'domain_id': note_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(note_vo.created_at)
        })

    return note_pb2.NoteInfo(**info)


def NotesInfo(note_vos, total_count, **kwargs):
    return note_pb2.NotesInfo(results=list(
        map(functools.partial(NoteInfo, **kwargs), note_vos)), total_count=total_count)



import logging
from datetime import datetime, timedelta
from spaceone.core.service import *
from spaceone.core import config
from spaceone.inventory.manager.identity_manager import IdentityManager
from spaceone.inventory.manager.cleanup_manager import CleanupManager
from spaceone.inventory.manager.cloud_service_manager import CloudServiceManager
from spaceone.inventory.manager.record_manager import RecordManager
from spaceone.inventory.manager.note_manager import NoteManager
from spaceone.inventory.manager.job_manager import JobManager
from spaceone.inventory.manager.job_task_manager import JobTaskManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@event_handler
class CleanupService(BaseService):

    @transaction
    @append_query_filter([])
    def list_domains(self, params):
        identity_mgr: IdentityManager = self.locator.get_manager(IdentityManager)
        return identity_mgr.list_domains(params.get('query', {}))

    @transaction
    @check_required(['domain_id'])
    def update_job_state(self, params):
        """
        Args:
            params (dict): {
                'options': 'dict',
                'domain_id': 'str'
            }

        Based on domain's cleanup policy update job.state

        Returns:
            None
        """
        _LOGGER.debug(f'[update_job_state] START')
        job_mgr: JobManager = self.locator.get_manager(JobManager)

        domain_id = params['domain_id']
        job_timeout = config.get_global('JOB_TIMEOUT', 2)  # hours

        policies = {'inventory.Job': {'TIMEOUT': job_timeout}}

        for resource_type, policy in policies.items():
            for status, hour in policy.items():
                _LOGGER.debug(f'[update_job_state] {resource_type}, {hour}, {status}, {domain_id}')
                job_mgr.update_job_timeout_by_hour(hour, status, domain_id)

    @transaction
    @check_required(['domain_id'])
    def terminate_jobs(self, params):
        """
        Args:
            params (dict): {
                'options': 'dict',
                'domain_id': 'str'
            }

        terminate old resources

        Returns:
            None
        """

        domain_id = params['domain_id']

        job_mgr: JobManager = self.locator.get_manager('JobManager')
        job_task_mgr: JobTaskManager = self.locator.get_manager('JobTaskManager')
        termination_time = config.get_global('JOB_TERMINATION_TIME', 30 * 2)  # days

        query = {
            'filter': [
                {
                    'k': 'created_at',
                    'v': datetime.utcnow() - timedelta(days=termination_time),
                    'o': 'lt'
                },
                {
                    'k': 'domain_id',
                    'v': domain_id,
                    'o': 'eq'
                }
            ]
        }

        job_vos, job_total_count = job_mgr.list_jobs(query)
        job_task_vos, job_task_total_count = job_task_mgr.list(query)

        job_vos.delete()
        job_task_vos.delete()

        _LOGGER.info(f'[terminate_jobs] Terminate jobs: {str(job_total_count)}')
        _LOGGER.info(f'[terminate_jobs] Terminate job tasks: {str(job_task_total_count)}')

    @transaction
    @check_required(['domain_id'])
    def delete_resources(self, params):
        """
        Args:
            params (dict): {
                'options': 'dict',
                'domain_id': 'str'
            }

        Based on domain's delete policy, delete resources

        Returns:
            None
        """

        domain_id = params['domain_id']
        exclude_domains = config.get_global('DELETE_EXCLUDE_DOMAINS', [])

        if domain_id not in exclude_domains:
            # Get Delete Policy of domain
            # TODO: from domain config
            # policies = self._get_domain_config(state, domain_id)

            policies = config.get_global('DEFAULT_DELETE_POLICIES', {})
            _LOGGER.debug(f'[delete_resources] {policies}')

            cleanup_mgr: CleanupManager = self.locator.get_manager(CleanupManager)
            for resource_type, hour in policies.items():
                try:
                    # _LOGGER.debug(f'[delete_resources] {resource_type}, {hour}, {domain_id}')
                    deleted_count = cleanup_mgr.delete_resources_by_policy(resource_type, hour, domain_id)
                    _LOGGER.debug(f'[delete_resources] number of deleted count: {deleted_count}')

                    # TODO: event notification
                except Exception as e:
                    _LOGGER.error(f'[delete_resources] {e}')
        else:
            _LOGGER.debug(f'[delete_resources] skip domain: {domain_id}')

    @transaction
    @check_required(['domain_id'])
    def terminate_resources(self, params):
        """
        Args:
            params (dict): {
                'options': 'dict',
                'domain_id': 'str'
            }

        Returns:
            None
        """
        cloud_svc_mgr: CloudServiceManager = self.locator.get_manager(CloudServiceManager)
        record_mgr: RecordManager = self.locator.get_manager(RecordManager)
        note_mgr: NoteManager = self.locator.get_manager(NoteManager)

        domain_id = params['domain_id']
        termination_time = config.get_global('RESOURCE_TERMINATION_TIME', 30 * 6)  # days
        _LOGGER.debug(f'[terminate_resources] RESOURCE_TERMINATION_TIME = {termination_time}')

        query = {
            'filter': [
                {'k': 'deleted_at', 'v': datetime.utcnow() - timedelta(days=termination_time), 'o': 'lt'},
                {'k': 'state', 'v': 'DELETED', 'o': 'eq'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'}
            ],
            'only': ['cloud_service_id']
        }
        # _LOGGER.debug(f'[terminate_resources] query: {query}')

        cloud_svc_vos, total_count = cloud_svc_mgr.list_cloud_services(query)
        _LOGGER.info(f'[terminate_resources] Terminate cloud services: {str(total_count)}')

        for cloud_svc_vo in cloud_svc_vos:
            cloud_service_id = cloud_svc_vo.cloud_service_id
            # _LOGGER.info(f'[terminate_resources] Terminate cloud service / record / note: {cloud_service_id}')

            # Cascade Delete Records
            record_vos = record_mgr.filter_records(cloud_service_id=cloud_service_id, domain_id=domain_id)
            record_vos.delete()

            # Cascade Delete Notes
            note_vos = note_mgr.filter_notes(cloud_service_id=cloud_service_id, domain_id=domain_id)
            note_vos.delete()

            cloud_svc_mgr.terminate_cloud_service(cloud_service_id, domain_id)



# -*- coding: utf-8 -*-
from spaceone.core.error import *


class ERROR_NOT_FOUND_USER_IN_REGION(ERROR_BASE):
    _message = 'A user "{user_id}" is not exist in region ({region_id}).'


class ERROR_ALREADY_EXIST_USER_IN_REGION(ERROR_BASE):
    _message = 'A user "{user_id}" is already exist in region ({region_id}).'



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class CloudServiceType(MongoModel):
    cloud_service_type_id = StringField(max_length=40, generate_id='cloud-svc-type', unique=True)
    name = StringField(max_length=255, unique_with=['provider', 'group', 'domain_id'])
    provider = StringField(max_length=255)
    group = StringField(max_length=255)
    cloud_service_type_key = StringField(max_length=255)
    ref_cloud_service_type = StringField(max_length=255)
    service_code = StringField(max_length=255, default=None, null=True)
    is_primary = BooleanField(default=False)
    is_major = BooleanField(default=False)
    resource_type = StringField(max_length=255)
    labels = ListField(StringField(max_length=255))
    metadata = DictField()
    tags = DictField()
    domain_id = StringField(max_length=40)
    updated_by = StringField(default=None, null=True)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        'updatable_fields': [
            'cloud_service_type_key',
            'service_code',
            'is_primary',
            'is_major',
            'resource_type',
            'metadata',
            'labels',
            'tags',
            'updated_by',
            'updated_at'
        ],
        'minimal_fields': [
            'cloud_service_type_id',
            'name',
            'provider',
            'group',
            'service_code',
            'is_primary',
            'is_major',
            'resource_type'
        ],
        'ordering': [
            'provider',
            'group',
            'name'
        ],
        'indexes': [
            {
                "fields": ['domain_id', 'cloud_service_type_id'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_1"
            },
            {
                "fields": ['domain_id', 'provider', 'group', 'name', 'is_primary'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_2"
            },
            {
                "fields": ['domain_id', 'cloud_service_type_key'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_3"
            },
            {
                "fields": ['cloud_service_type_id', 'ref_cloud_service_type'],
                "name": "COMPOUND_INDEX_FOR_REF_1"
            },
            {
                "fields": ['labels', 'is_primary', 'ref_cloud_service_type'],
                "name": "COMPOUND_INDEX_FOR_REF_2"
            },
        ]
    }



import logging
from datetime import datetime, timedelta
from spaceone.core.manager import BaseManager
from spaceone.inventory.model.job_model import Job
from spaceone.inventory.error import *
from spaceone.inventory.lib.job_state import JobStateMachine
from spaceone.inventory.conf.collector_conf import *

_LOGGER = logging.getLogger(__name__)


class JobManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.job_model: Job = self.locator.get_model('Job')

    def create_job(self, collector_vo, params):
        """ Create Job for collect method
        Args:
            collector_vo: collector model
            params(dict): {
                'collector_id': str,
                'secret_id': str,
                'domain_id': str
            }
        Returns: job_vo
        """
        job_params = params.copy()
        job_params['collector'] = collector_vo
        return self.job_model.create(job_params)

    def update(self, job_id, params, domain_id):
        job_vo = self.get_job(job_id, domain_id)
        return self.update_job_by_vo(params, job_vo)

    def delete(self, job_id, domain_id):
        job_vo = self.get_job(job_id, domain_id)
        job_vo.delete()

    def get_job(self, job_id, domain_id, only=None):
        return self.job_model.get(job_id=job_id, domain_id=domain_id, only=only)

    def list_jobs(self, query):
        return self.job_model.query(**query)

    def analyze_jobs(self, query):
        return self.job_model.analyze(**query)

    def stat_jobs(self, query):
        return self.job_model.stat(**query)

    def increase_success_tasks(self, job_id, domain_id):
        job_vo: Job = self.get_job(job_id, domain_id)
        return self.increase_success_tasks_by_vo(job_vo)

    def increase_failure_tasks(self, job_id, domain_id):
        job_vo: Job = self.get_job(job_id, domain_id)
        return self.increase_failure_tasks_by_vo(job_vo)

    def decrease_remained_tasks(self, job_id, domain_id):
        job_vo: Job = self.get_job(job_id, domain_id)
        return self.decrease_remained_tasks_by_vo(job_vo)

    def decrease_remained_tasks_by_vo(self, job_vo: Job):
        job_vo = job_vo.decrement('remained_tasks')

        if job_vo.remained_tasks == 0 and job_vo.status != 'CANCELED':
            if job_vo.mark_error:
                self.make_failure_by_vo(job_vo)
            else:
                self.make_success_by_vo(job_vo)

        if job_vo.remained_tasks < 0:
            _LOGGER.debug(f'[decrease_remained_tasks] {job_vo.job_id}, {job_vo.remained_tasks}')
            raise ERROR_JOB_UPDATE(param='remained_tasks')

        return job_vo

    def add_error(self, job_id, domain_id, error_code, msg, additional=None):
        """
        error_info (dict): {
            'error_code': str,
            'message': str,
            'additional': dict
        }
        """
        message = repr(msg)
        error_info = {'error_code': error_code, 'message': message[:MAX_MESSAGE_LENGTH]}

        if additional:
            error_info['additional'] = additional

        job_vo = self.get_job(job_id, domain_id)
        job_vo.append('errors', error_info)
        self.mark_error_by_vo(job_vo)
        return job_vo

    def update_job_timeout_by_hour(self, hour, status, domain_id):
        created_at = datetime.utcnow() - timedelta(hours=hour)
        query = {
            'filter': [
                {'k': 'created_at', 'v': created_at, 'o': 'lt'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
                {'k': 'status', 'v': ['CREATED', 'IN_PROGRESS'], 'o': 'in'}
            ]
        }

        jobs, total_count = self.list_jobs(query)
        for job in jobs:
            self.make_timeout_by_vo(job)

    def list_duplicate_jobs(self, collector_id, secret_id, domain_id):
        # started_at = datetime.utcnow() - timedelta(minutes=10)

        query = {
            'filter': [
                {'k': 'collector_id', 'v': collector_id, 'o': 'eq'},
                {'k': 'secret_id', 'v': secret_id, 'o': 'eq'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
                {'k': 'status', 'v': 'IN_PROGRESS', 'o': 'eq'},
                # {'k': 'started_at', 'v': started_at, 'o': 'gte'},
            ]
        }

        job_vos, total_count = self.list_jobs(query)
        return job_vos

    def make_inprogress_by_vo(self, job_vo):
        job_state_machine = JobStateMachine(job_vo)
        job_state_machine.inprogress()
        self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    def make_success_by_vo(self, job_vo):
        job_state_machine = JobStateMachine(job_vo)
        job_state_machine.success()
        self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    def make_failure_by_vo(self, job_vo):
        job_state_machine = JobStateMachine(job_vo)
        job_state_machine.failure()
        self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    def make_canceled_by_vo(self, job_vo):
        job_state_machine = JobStateMachine(job_vo)
        job_state_machine.canceled()
        self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    # DEPRECATED
    # def make_timeout_by_vo(self, job_vo):
    #     job_state_machine = JobStateMachine(job_vo)
    #     job_state_machine.timeout()
    #     self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    def make_inprogress(self, job_id, domain_id):
        job_vo = self.get_job(job_id, domain_id)
        job_state_machine = JobStateMachine(job_vo)
        job_state_machine.inprogress()
        self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    def make_success(self, job_id, domain_id):
        job_vo = self.get_job(job_id, domain_id)
        job_state_machine = JobStateMachine(job_vo)
        job_state_machine.success()
        self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    def make_canceled(self, job_id, domain_id):
        job_vo = self.get_job(job_id, domain_id)
        job_state_machine = JobStateMachine(job_vo)
        job_state_machine.canceled()
        self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    def make_failure(self, job_id, domain_id):
        job_vo = self.get_job(job_id, domain_id)
        job_state_machine = JobStateMachine(job_vo)
        job_state_machine.failure()
        self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    # DEPRECATED
    # def make_timeout(self, job_id, domain_id):
    #     job_vo = self.get_job(job_id, domain_id)
    #     job_state_machine = JobStateMachine(job_vo)
    #     job_state_machine.timeout()
    #     self._update_job_status_by_vo(job_vo, job_state_machine.get_status())

    def is_canceled(self, job_id, domain_id):
        job_vo = self.get_job(job_id, domain_id)
        job_state_machine = JobStateMachine(job_vo)
        if job_state_machine.get_status() == 'CANCELED':
            return True
        return False

    def check_cancel(self, job_id, domain_id):
        job_vo = self.get_job(job_id, domain_id)
        job_state_machine = JobStateMachine(job_vo)
        job_status = job_state_machine.get_status()

        if job_status == 'CANCELED':
            return True
        return False

    def mark_error(self, job_id, domain_id):
        job_vo = self.get_job(job_id, domain_id)
        job_vo.update({'mark_error': 1})

    @staticmethod
    def mark_error_by_vo(job_vo):
        job_vo.update({'mark_error': 1})

    @staticmethod
    def delete_job_by_vo(job_vo):
        job_vo.delete()

    @staticmethod
    def _update_job_status_by_vo(job_vo, status):
        params = {'status': status}
        if status in ['SUCCESS', 'FAILURE', 'CANCELED']:
            params.update({'finished_at': datetime.utcnow()})

        _LOGGER.debug(f'[update_job_status] job_id: {job_vo.job_id}, status: {status}')
        return job_vo.update(params)

    @staticmethod
    def update_job_by_vo(params, job_vo: Job):
        return job_vo.update(params)

    @staticmethod
    def increase_success_tasks_by_vo(job_vo: Job):
        return job_vo.increment('success_tasks')

    @staticmethod
    def increase_failure_tasks_by_vo(job_vo: Job):
        return job_vo.increment('failure_tasks')



"""
DEPRECATED (NOT USED)
"""
import json
import logging
from spaceone.core.scheduler.worker import BaseWorker
from spaceone.core.locator import Locator
from spaceone.core.transaction import Transaction
from spaceone.core import queue
from spaceone.inventory.manager.collecting_manager import CollectingManager


_LOGGER = logging.getLogger(__name__)

class InventoryDBUpdater(BaseWorker):
    def __init__(self, queue, **kwargs):
        BaseWorker.__init__(self, queue, **kwargs)
        self.locator = Locator()

    def run(self):
        collecting_mgr: CollectingManager = self.locator.get_manager(CollectingManager)

        while True:
            try:
                resource_info = json.loads(queue.get(self.queue).decode())
                collecting_mgr.transaction = Transaction(resource_info['meta'])

                method = resource_info['method']
                if method == 'check_resource_state':
                    collecting_mgr.check_resource_state(resource_info['res'], resource_info['param'])
                elif method == 'watchdog_job_task_stat':
                    collecting_mgr.watchdog_job_task_stat(resource_info['param'])
                else:
                    _LOGGER.error(f'Unknown request: {resource_info}')

            except Exception as e:
                _LOGGER.error(f'[{self._name_}] failed to processing: {e}')
                continue




import logging
import json
from jsonschema import validate
from datetime import datetime, timedelta
from spaceone.core import config, queue
from spaceone.core.token import get_token
from spaceone.core.manager import BaseManager
from spaceone.core.scheduler.task_schema import SPACEONE_TASK_SCHEMA
from spaceone.inventory.manager.job_manager import JobManager
from spaceone.inventory.model.job_task_model import JobTask
from spaceone.inventory.lib.job_task_state import JobTaskStateMachine

_LOGGER = logging.getLogger(__name__)


class JobTaskManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.job_task_model: JobTask = self.locator.get_model('JobTask')

    def create_job_task(self, job_vo, domain_id, task_options):
        def _rollback(job_task_vo):
            _LOGGER.info(f'[ROLLBACK] Delete job_task: {job_task_vo.job_task_id}')
            job_task_vo.delete()

        params = {
            'job_id': job_vo.job_id,
            'collector_id': job_vo.collector_id,
            'domain_id': domain_id,
            'options': task_options
        }

        job_task_vo: JobTask = self.job_task_model.create(params)
        self.transaction.add_rollback(_rollback, job_task_vo)
        return job_task_vo

    def get(self, job_task_id, domain_id, only=None):
        return self.job_task_model.get(job_task_id=job_task_id, domain_id=domain_id, only=only)

    def list(self, query):
        return self.job_task_model.query(**query)

    def stat(self, query):
        return self.job_task_model.stat(**query)

    def delete(self, job_task_id, domain_id):
        job_task_vo = self.get(job_task_id, domain_id)
        job_task_vo.delete()

    def push_job_task(self, params):
        task = self.create_task_pipeline(params)
        validate(task, schema=SPACEONE_TASK_SCHEMA)
        json_task = json.dumps(task)
        queue.put(self.get_queue_name(name='collect_queue'), json_task)

    def check_duplicate_job_tasks(self, collector_id, secret_id, domain_id):
        started_at = datetime.utcnow() - timedelta(minutes=10)

        query = {
            'filter': [
                {'k': 'collector_id', 'v': collector_id, 'o': 'eq'},
                {'k': 'secret_id', 'v': secret_id, 'o': 'eq'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
                {'k': 'status', 'v': 'IN_PROGRESS', 'o': 'eq'},
                {'k': 'started_at', 'v': started_at, 'o': 'gte'},
            ]
        }

        job_task_vos, total_count = self.list(query)

        if total_count > 0:
            for job_task_vo in job_task_vos:
                _LOGGER.debug(f'[check_duplicate_job_tasks] Duplicate Job Info: {job_task_vo.job_id} '
                              f'({job_task_vo.collector_id})')
            return True

        return False

    def add_error(self, job_task_id, domain_id, error_code, msg, additional=None):
        job_mgr: JobManager = self.locator.get_manager(JobManager)

        error_info = {'error_code': error_code, 'message': str(msg).strip()}

        if additional:
            error_info['additional'] = additional

        job_task_vo = self.get(job_task_id, domain_id)
        job_task_vo.append('errors', error_info)
        job_mgr.mark_error(job_task_vo.job_id, domain_id)
        _LOGGER.error(f'[add_error] {job_task_id}: {error_info}', exc_info=True)

        return job_task_vo

    def _update_job_status(self, job_task_id, status, domain_id, started_at=None, finished_at=None, secret_info=None, collecting_count_info=None):
        job_task_vo = self.get(job_task_id, domain_id)
        params = {'status': status}

        if started_at:
            params['started_at'] = started_at

        if finished_at:
            params['finished_at'] = finished_at

        if secret_info:
            params.update(secret_info)

        if collecting_count_info:
            params.update(collecting_count_info)

        _LOGGER.debug(f'[update_job_status] job_task_id: {job_task_id}, status: {status}')
        return job_task_vo.update(params)

    def make_inprogress(self, job_task_id, domain_id, secret_info=None, collecting_count_info=None):
        job_task_vo = self.get(job_task_id, domain_id)
        job_state_machine = JobTaskStateMachine(job_task_vo)
        job_state_machine.inprogress()
        self._update_job_status(job_task_id, job_state_machine.get_state(), domain_id,
                                started_at=datetime.utcnow(),
                                secret_info=secret_info,
                                collecting_count_info=collecting_count_info)

    def make_success(self, job_task_id, domain_id, secret_info=None, collecting_count_info=None):
        job_task_vo = self.get(job_task_id, domain_id)
        job_state_machine = JobTaskStateMachine(job_task_vo)
        job_state_machine.success()
        self._update_job_status(job_task_id, job_state_machine.get_state(), domain_id,
                                finished_at=datetime.utcnow(),
                                secret_info=secret_info,
                                collecting_count_info=collecting_count_info)

    def make_failure(self, job_task_id, domain_id, secret_info=None, collecting_count_info=None):
        job_task_vo = self.get(job_task_id, domain_id)
        job_state_machine = JobTaskStateMachine(job_task_vo)
        job_state_machine.failure()
        self._update_job_status(job_task_id, job_state_machine.get_state(), domain_id,
                                finished_at=datetime.utcnow(),
                                secret_info=secret_info,
                                collecting_count_info=collecting_count_info)

    def make_canceled(self, job_task_id, domain_id, secret_info=None, collecting_count_info=None):
        job_task_vo = self.get(job_task_id, domain_id)
        job_state_machine = JobTaskStateMachine(job_task_vo)
        job_state_machine.canceled()
        self._update_job_status(job_task_id, job_state_machine.get_state(), domain_id,
                                finished_at=datetime.utcnow(),
                                secret_info=secret_info,
                                collecting_count_info=collecting_count_info)

    @staticmethod
    def delete_job_task_by_vo(job_task_vo):
        job_task_vo.delete()

    @staticmethod
    def get_queue_name(name='collect_queue'):
        try:
            return config.get_global(name)
        except Exception as e:
            _LOGGER.warning(f'[_get_queue_name] name: {name} is not configured')
            return None

    @staticmethod
    def create_task_pipeline(params):
        task = {
            'locator': 'MANAGER',
            'name': 'CollectingManager',
            'metadata': {'token': get_token(), 'domain_id': params.get('domain_id')},
            'method': 'collecting_resources',
            'params': {'params': params}
        }

        stp = {
            'name': 'collecting_resources',
            'version': 'v1',
            'executionEngine': 'BaseWorker',
            'stages': [task]
        }
        # _LOGGER.debug(f'[_create_task] tasks: {stp}')
        return stp
